
#include <net-snmp/net-snmp-config.h>
#include <net-snmp/net-snmp-includes.h>
#include <net-snmp/agent/net-snmp-agent-includes.h>
#include "../../vtk.h"
#include "dot1dStatic.h"
#include "../../net_snmp_handle.h"

oid dot1dStatic_variables_oid[] = { SNMP_OID_MIB2, 17, 5 };

struct variable4 dot1dStatic_variables[] = {
    {DOT1DSTATICADDRESS,       ASN_OCTET_STR, RWRITE, var_dot1dStatic, 3, { 1, 1, 1 }},
    {DOT1DSTATICRECEIVEPORT,   ASN_INTEGER,   RWRITE, var_dot1dStatic, 3, { 1, 1, 2 }},
    {DOT1DSTATICALLOWEDTOGOTO, ASN_OCTET_STR, RWRITE, var_dot1dStatic, 3, { 1, 1, 3 }},   
    {DOT1DSTATICSTATUS,        ASN_INTEGER,   RWRITE, var_dot1dStatic, 3, { 1, 1, 4 }},
};

static dot1dStaticMACFTEntry_rowNumber = 0;

static struct MIB_dot1dMACFilterTableEntry {
    unsigned char   static_address[ETH_MAC_LEN];
    long            recv_port;
    unsigned char   allowtogo_address[ETH_MAC_LEN];
    long            status;
} dot1dMACFTEntry[MAX_dot1dStaticMACFTEntry_rowNumber];

void init_dot1dStaticMACFTPortEntry(void);

void init_dot1dStatic(void)
{
    DEBUGMSGTL(("dot1dStatic", "Initializing\n"));

    REGISTER_MIB("dot1dStatic",dot1dStatic_variables,variable4,dot1dStatic_variables_oid);
    init_dot1dStaticMACFTPortEntry();
}

void append_dot1dStaticMACFTEntry( unsigned char *static_address,
                                   long recv_port, unsigned char *allowtogo_address,
                                   long status )
{
    int row;

    if( dot1dStaticMACFTEntry_rowNumber == MAX_dot1dStaticMACFTEntry_rowNumber )
        return;

    row = dot1dStaticMACFTEntry_rowNumber;
    memset( (char*)&dot1dMACFTEntry[row], 0, sizeof(struct MIB_dot1dMACFilterTableEntry) ); //[init]

    dot1dMACFTEntry[row].recv_port = recv_port;
    dot1dMACFTEntry[row].status = status;

    memcpy(dot1dMACFTEntry[row].static_address, static_address, ETH_MAC_LEN);
    memcpy(dot1dMACFTEntry[row].allowtogo_address, allowtogo_address, ETH_MAC_LEN);

    dot1dStaticMACFTEntry_rowNumber++;
}

void init_dot1dStaticMACFTPortEntry(void)
{
    int     index;
    long    recv_port = 0;
    long    status = 1;;
    unsigned char mac[ETH_MAC_LEN + 1];

    memset((u_char*)&mac[0], '\0', ETH_MAC_LEN + 1);
    mac[0]=0xFF;
    mac[1]=0xFF;
    mac[2]=0xFF;
    mac[3]=0xFF;
    mac[4]=0xFF;
    mac[5]=0x00;
    for( index=0 ; index < 3 ; index++ )
    {
        append_dot1dStaticMACFTEntry(mac, recv_port, mac, status);
        mac[5]++;
    }
}

struct MIB_dot1dMACFilterTableEntry *get_dot1dStaticMACFTEntry(unsigned char *string)
{
    long            row;
    unsigned char   *curr_str;

    curr_str = string;
    for( row=0 ; row < dot1dStaticMACFTEntry_rowNumber ; row++)
    {
        if( curr_str[0] == dot1dMACFTEntry[row].static_address[0] &&
            curr_str[1] == dot1dMACFTEntry[row].static_address[1] &&
            curr_str[2] == dot1dMACFTEntry[row].static_address[2] &&
            curr_str[3] == dot1dMACFTEntry[row].static_address[3] &&
            curr_str[4] == dot1dMACFTEntry[row].static_address[4] &&
            curr_str[5] == dot1dMACFTEntry[row].static_address[5] )
            break;
    }

    if( row == dot1dStaticMACFTEntry_rowNumber )
        return NULL;
        
    return (struct MIB_dot1dMACFilterTableEntry *)&dot1dMACFTEntry[row];
}

struct MIB_dot1dMACFilterTableEntry *getnext_dot1dStaticMACFTEntry(unsigned char *string)
{
    long            row;
    unsigned char   *curr_str;
	
    curr_str = string;
    for( row=0 ; row < dot1dStaticMACFTEntry_rowNumber ; row++ )
    {
        if( curr_str[0]<=dot1dMACFTEntry[row].static_address[0] )
            if( curr_str[0]<dot1dMACFTEntry[row].static_address[0] )
                break;
            else
                if( curr_str[1]<=dot1dMACFTEntry[row].static_address[1] )
                    if( curr_str[1]<dot1dMACFTEntry[row].static_address[1] )
                        break;
                    else
                        if( curr_str[2]<=dot1dMACFTEntry[row].static_address[2] )
                            if( curr_str[2]<dot1dMACFTEntry[row].static_address[2] )
                                break;
                            else
                                if( curr_str[3]<=dot1dMACFTEntry[row].static_address[3] )
                                    if( curr_str[3]<dot1dMACFTEntry[row].static_address[3] )
                                        break;
                                    else
                                        if( curr_str[4]<=dot1dMACFTEntry[row].static_address[4] )
                                            if( curr_str[4]<dot1dMACFTEntry[row].static_address[4] )
                                                break;
                                            else
                                                if( curr_str[5]<dot1dMACFTEntry[row].static_address[5] )
                                                    break;
    }

    if( row == dot1dStaticMACFTEntry_rowNumber )
        return NULL;

    return (struct MIB_dot1dMACFilterTableEntry *)&dot1dMACFTEntry[row];
}

unsigned char *var_dot1dStatic( struct variable *vp, oid *name,
                                size_t *length, int exact,
                                size_t *var_len, WriteMethod **write_method )
{
    static long             long_ret;
    static unsigned char    string[SPRINT_MAX_LEN];
    u_long                  i;
    snmp_eventMsg_t         event;
      
    *write_method = NULL;

    if( memcmp(name, vp->name, sizeof(oid)*vp->namelen) != 0 )
    {
        memcpy(name, vp->name, sizeof(oid)*vp->namelen);
        *length = vp->namelen;
    }

    if( exact )
    {
        if( *length != 17 )
            return NULL;

        for( i=0 ; i<6 ; i++ )
            string[i] = name[11 + i];
        string[i] = '\0';
    } else
        if( *length > 16 )
        {
            for( i=0 ; i<6 ; i++ )
                string[i] = name[11 + i];
            string[i] = '\0';
        }
        else
            memset(string, '\0', 7);

return NULL;    /* Not support this table! */
/* @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@  */

    *var_len = sizeof(long_ret);
    memset( &event, 0, sizeof(snmp_eventMsg_t) );
    memcpy(event.string,string, ETH_MAC_LEN);
    switch( vp->magic )
    {
        case DOT1DSTATICADDRESS:
            event.type=e_snmp_dot1dStaticAddress_get;
            util_snmp_evtMsg_ack_handler( &event );
            memcpy(string, event.string, ETH_MAC_LEN);
            *var_len = ETH_MAC_LEN;
            return (u_char*) &string;

        case DOT1DSTATICRECEIVEPORT:
            event.type=e_snmp_dot1dStaticReceivePort_get;
            util_snmp_evtMsg_ack_handler( &event );
            long_ret = event.val;
            *var_len = sizeof(long_ret);
            return (u_char *)&long_ret;

        case DOT1DSTATICALLOWEDTOGOTO:
            event.type=e_snmp_dot1dStaticAllowedToGoTo_get;
            util_snmp_evtMsg_ack_handler( &event );
            memcpy(string, event.string, 4);
            *var_len = 4;
            return (u_char*) &string;

        case DOT1DSTATICSTATUS:
            event.type=e_snmp_dot1dStaticStatus_get;
            util_snmp_evtMsg_ack_handler( &event );
            long_ret = event.val;
            *var_len = sizeof(long_ret);
            return (u_char *)&long_ret;

        default:
            ERROR_MSG("var_dot1dStatic() Fail");
    }
    return NULL;
}
