
#include <net-snmp/net-snmp-config.h>
#include <net-snmp/net-snmp-includes.h>
#include <net-snmp/agent/net-snmp-agent-includes.h>

#include "dot3CollTable.h"

oid dot3CollTable_variables_oid[] = { SNMP_OID_MIB2, 10, 7, 5 };

struct variable4 dot3CollTable_variables[] = {
    {DOT3COLLFREQUENCIES,   ASN_INTEGER,    RONLY,  var_dot3CollTable, 2,   { 1, 3 }},
};


static dot3CollTableTEntry_rowNumber = 0;


static struct MIB_dot3CollTableTEntry {
    long port;
    long frequence;
} dot3CollTableTEntry[MAX_dot3CollTableTEntry_rowNumber];


void init_dot3CollTableEntry(void);

struct MIB_dot3CollTableTEntry *get_dot3CollTableEntry(long port);
struct MIB_dot3CollTableTEntry *getnext_dot3CollTableEntry(long port);

int append_dot3CollTableEntry(long port, long frequence);


void init_dot3CollTable(void)
{
    DEBUGMSGTL(("dot1dStatic", "Initializing\n"));

    REGISTER_MIB("dot1dStatic", 
        dot3CollTable_variables, variable4, dot3CollTable_variables_oid);

    init_dot3CollTableEntry();
}

void init_dot3CollTableEntry(void)
{
    long port;
    long frequence = 0;
 
    for (port = 0; port < MAX_dot3CollTableTEntry_rowNumber; port++)
        append_dot3CollTableEntry(port + 1, frequence);
}

int append_dot3CollTableEntry(long port, long frequence)
{
    int row;
    if (dot3CollTableTEntry_rowNumber == MAX_dot3CollTableTEntry_rowNumber)
        return -1;
   
    row = dot3CollTableTEntry_rowNumber;   
    memset( (char*)&dot3CollTableTEntry[row], 0, sizeof(struct MIB_dot3CollTableTEntry) ); //[init]
   	
    dot3CollTableTEntry[row].port = port;
    dot3CollTableTEntry[row].frequence = frequence;
    
    dot3CollTableTEntry_rowNumber++;
    return 0;
}		

struct MIB_dot3CollTableTEntry *get_dot3CollTableEntry(long port)
{
    long curr_index_value, row;

    curr_index_value = port;
    for (row = 0; row < dot3CollTableTEntry_rowNumber; row++) {
        if (curr_index_value == dot3CollTableTEntry[row].port)
            break;
    }
	
    if (row == dot3CollTableTEntry_rowNumber) {
        return NULL;
    } 
    return (struct MIB_dot3CollTableTEntry *)&dot3CollTableTEntry[row]; 
}

struct MIB_dot3CollTableTEntry *getnext_dot3CollTableEntry(long port)
{
    long curr_index_value, row;

    curr_index_value = port;
    for (row = 0; row < dot3CollTableTEntry_rowNumber; row++) {
        if (curr_index_value < dot3CollTableTEntry[row].port)
            break;
    }

    if (row == dot3CollTableTEntry_rowNumber) {
        return NULL;
    } 
    return (struct MIB_dot3CollTableTEntry*)&dot3CollTableTEntry[row]; 
}

unsigned char *var_dot3CollTable(struct variable *vp, oid *name, 
    size_t *length, int exact, size_t *var_len, WriteMethod **write_method)
{
    long index;
    static long long_ret;
    struct MIB_dot3CollTableTEntry *geptr;
    
    *write_method = NULL;

    if (memcmp(name, vp->name, sizeof(oid)*vp->namelen) != 0) {
        memcpy(name, vp->name, sizeof(oid)*vp->namelen);
        *length = vp->namelen;
    }

    // 1,3,6,1,2,1,10,7,5,1,item
    //                      11 
    // 1,3,6,1,2,1,10,7,5,1,item,[index]
    //                              12 
    if (exact) {
        if (*length != 12)
            return NULL;
        
        index = name[11];
        geptr = get_dot3CollTableEntry(index);
    } else {
        index = *length > 11 ? name[11] : 0;

        geptr = getnext_dot3CollTableEntry(index);
        if (geptr) { 
    	    name[11] = geptr->port;
    	    *length = 12;
        } 	
    }	

    if (geptr == NULL)
        return NULL;

    *var_len = sizeof(long_ret);

    switch(vp->magic) {
        case DOT3COLLFREQUENCIES:
            long_ret = geptr->frequence;
            return (u_char*) &long_ret;

        default:                                                         
            ERROR_MSG("var_dot3CollTable() Fail");
    }
    return NULL;
}

