
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <asm/param.h>
#include <net/if.h>
#include <sys/socket.h> 
#include <netinet/in.h> 
#include <fcntl.h> 
#include <sys/select.h>  
#include "../../vtk.h"
#include "util_if.h"
#include "../../net_snmp_handle.h"

/*-----------------------------------------------------------------------------
 * FUNCTION NAME: util_snmp_evtMsg_handler
 * PURPOSE: This function provide to handler/parse SNMP command socket.
 *          Send command to CLI, No Ack!
 * PARAMs:
 *		INPUT:  snmp_eventMsg_t *event
 *              
 *		OUTPUT: None.
 *
 * RETURN:	int -- 0: success/ else: failure
 *
 * NOTES: 
 *-----------------------------------------------------------------------------
 */
int util_snmp_evtMsg_handler(snmp_eventMsg_t *event)
{
    int                 socket_id, len;
    struct sockaddr_in  cli_addr; 

    socket_id = socket(AF_INET, SOCK_DGRAM, 0); 
    if( socket_id < 0 ) 
    {
        printf( "%s socket open fail!!(%d) \n", __func__, socket_id);
        return -1; 
    }

    bzero(&cli_addr, sizeof(cli_addr));
    cli_addr.sin_family      = AF_INET; 
    cli_addr.sin_addr.s_addr = inet_addr(LOCAL_ADDRESS); 
    cli_addr.sin_port        = htons(VTK_SNMP_SERVER_2_PORT);
    len = sendto( socket_id, event, sizeof(snmp_eventMsg_t),0,
                 (struct sockaddr *)&cli_addr,sizeof(cli_addr));
    if( len<0 )
    {
        close(socket_id);
        return -2; 
    }
    close(socket_id);
	return 0;
}

/*-----------------------------------------------------------------------------
 * FUNCTION NAME: util_snmp_evtMsg_ack_handler
 * PURPOSE: This function provide to handler send command to CLI, Waiting for Ack!
 *          Send command to CLI, No Ack!
 * PARAMs:
 *		INPUT:  snmp_eventMsg_t *event
 *              
 *		OUTPUT: None.
 *
 * RETURN:	int -- 0: success/ else: failure
 *
 * NOTES: 
 *-----------------------------------------------------------------------------
 */
int util_snmp_evtMsg_ack_handler( snmp_eventMsg_t *event )
{
    int                 socket_id, len;
    struct sockaddr_in  cli_addr; 
    struct timeval      timeout;
    fd_set              fds;

    socket_id = socket(AF_INET, SOCK_DGRAM, 0); 
    if( socket_id < 0 ) 
    {
        printf( "%s socket open fail!!(%d) \n", __func__, socket_id);
        return -1; 
    }

    bzero(&cli_addr, sizeof(cli_addr));
    cli_addr.sin_family      = AF_INET; 
    cli_addr.sin_addr.s_addr = inet_addr(LOCAL_ADDRESS); 
    cli_addr.sin_port        = htons(VTK_SNMP_SERVER_2_PORT);

    len = sendto( socket_id, event, sizeof(snmp_eventMsg_t),0,
                  (struct sockaddr *)&cli_addr,sizeof(cli_addr));
    if( len<0 )
    {
        close(socket_id);
        return -2; 
    }

    timeout.tv_sec = 1;
    timeout.tv_usec= 0;
    FD_ZERO(&fds);
    FD_SET(socket_id, &fds);
    if( select(socket_id+1, &fds, 0, 0, &timeout) <= 0 )
    {
        close(socket_id);
        return -3;
    }
    else
    {
        len = recvfrom(socket_id, event, sizeof(snmp_eventMsg_t), 0, 0, 0);
        if( len<0 )
        {
            printf( "%s socket recvfrom fail!\n", __func__ );
            close(socket_id);
            event->val=0;
            return -4;
        }
    }
    close(socket_id);
    return RC_OK;
}
/*------------------------------End of Program--------------------------------*/
