
#include <net-snmp/net-snmp-config.h>
#include <net-snmp/net-snmp-includes.h>
#include <net-snmp/agent/net-snmp-agent-includes.h>

#include "../../vtk.h"
#include "interfacesIss.h"
#include "../lib/util_if.h"
#include "../../net_snmp_handle.h"

oid interfacesIss_variables_oid[] = { SNMP_OID_MIB2, 2 };

struct variable interfacesIss_variables[] = {
    {IFIssNUMBER,           ASN_INTEGER,    RONLY ,  var_interfacesIss, 1,  { 1 }},
    {IFIssINDEX,            ASN_INTEGER,    RONLY,   var_ifIssTable, 3,  { 2, 1, 1 }},
    {IFIssDESCR,            ASN_OCTET_STR,  RONLY,   var_ifIssTable, 3,  { 2, 1, 2 }},
    {IFIssTYPE,             ASN_INTEGER,    RONLY,   var_ifIssTable, 3,  { 2, 1, 3 }},
    {IFIssMTU,              ASN_INTEGER,    RONLY,   var_ifIssTable, 3,  { 2, 1, 4 }},
    {IFIssSPEED,            ASN_GAUGE,      RONLY,   var_ifIssTable, 3,  { 2, 1, 5 }},
    {IFIssPHYSADDRESS,      ASN_OCTET_STR,  RONLY,   var_ifIssTable, 3,  { 2, 1, 6 }},
    {IFIssADMINSTATUS,      ASN_INTEGER,    RWRITE,  var_ifIssTable, 3,  { 2, 1, 7 }},
    {IFIssOPERSTATUS,       ASN_INTEGER,    RONLY,   var_ifIssTable, 3,  { 2, 1, 8 }},
    {IFIssLASTCHANGE,       ASN_TIMETICKS,  RONLY,   var_ifIssTable, 3,  { 2, 1, 9 }},
    {IFIssINOCTETS,         ASN_COUNTER,    RONLY,   var_ifIssTable, 3,  { 2, 1, 10 }},
    {IFIssINUCASTPKTS,      ASN_COUNTER,    RONLY,   var_ifIssTable, 3,  { 2, 1, 11 }},
    {IFIssINNUCASTPKTS,     ASN_COUNTER,    RONLY,   var_ifIssTable, 3,  { 2, 1, 12 }},
    {IFIssINDISCARDS,       ASN_COUNTER,    RONLY,   var_ifIssTable, 3,  { 2, 1, 13 }},
    {IFIssINERRORS,         ASN_COUNTER,    RONLY,   var_ifIssTable, 3,  { 2, 1, 14 }},
    {IFIssINUNKNOWNPROTOS,  ASN_COUNTER,    RONLY,   var_ifIssTable, 3,  { 2, 1, 15 }},
    {IFIssOUTOCTETS,        ASN_COUNTER,    RONLY,   var_ifIssTable, 3,  { 2, 1, 16 }},
    {IFIssOUTUCASTPKTS,     ASN_COUNTER,    RONLY,   var_ifIssTable, 3,  { 2, 1, 17 }},
    {IFIssOUTNUCASTPKTS,    ASN_COUNTER,    RONLY,   var_ifIssTable, 3,  { 2, 1, 18 }},
    {IFIssOUTDISCARDS,      ASN_COUNTER,    RONLY,   var_ifIssTable, 3,  { 2, 1, 19 }},
    {IFIssOUTERRORS,        ASN_COUNTER,    RONLY,   var_ifIssTable, 3,  { 2, 1, 20 }},
    {IFIssOUTQLEN,          ASN_GAUGE,      RONLY,   var_ifIssTable, 3,  { 2, 1, 21 }},
    {IFIssSPECIFIC,         ASN_OBJECT_ID,  RONLY,   var_ifIssTable, 3,  { 2, 1, 22 }},
};

static int ifIssEntry_rowNumber = 0;
static time_t lastChanged = 0;
   
static struct MIB_ifIssScalar {
    long ifIssNumber;
} ifIssScalar;

static struct MIB_ifIssEntry{
    long index;
    char descr[SPRINT_MAX_LEN]; 
    long descr_len;
    long type;
    long mtu;
    u_long speed;
    char physAddress[SPRINT_MAX_LEN]; 
    long physAddress_len;
    long adminStatus;
    long operStatus;
    u_long lastChange;
    u_long inOctets;
    u_long inUCastPkts;
    u_long inNUCastPkts;
    u_long inDiscards;
    u_long inErrors;
    u_long inUnknownProtos;
    u_long outOctets;
    u_long outUCastPkts;
    u_long outNUCastPkts;
    u_long outDiscards;
    u_long outErrors;
    u_long outQLen;
    oid  specific[MAX_OID_LEN];
    long specific_len;   
} ifIssEntry[MAX_ifIssEntry_rowNumber+1];
   
void init_ifIssScalar(void);
void init_ifIssEntry(void);

struct MIB_ifIssEntry *get_ifIssEntry(u_long index);
struct MIB_ifIssEntry *getnext_ifIssEntry(u_long index);

int append_ifIssEntry(long index,  char *descr,  long type,  long mtu,  u_long speed,
    char *physAddress, long adminStatus, long operStatus, u_long lastChange,
    u_long inOctets, u_long inUCastPkts, u_long inNUCastPkts,
    u_long inDiscards, u_long inErrors, u_long inUnknownProtos,
    u_long outOctets, u_long outUCastPkts, u_long outNUCastPkts,
    u_long outDiscards, u_long outErrors, u_long outQLen, char  *specific
);

WriteMethod write_ifIssAdminStatus;
WriteMethod write_noSuchName;

void init_interfacesIss(void)
{
    REGISTER_MIB("interfacesIss",interfacesIss_variables, variable, interfacesIss_variables_oid);
    init_ifIssScalar();
    init_ifIssEntry();
}

void init_ifIssScalar(void)
{
    ifIssScalar.ifIssNumber = MAX_ifIssEntry_rowNumber;
}

void interfaceIss_port_description_set( int port, char *description )
{
    strcpy( ifIssEntry[port-1].descr, description );
}

void interfaceIss_port_description_get( int port, char *description )
{
    strcpy( description, ifIssEntry[port-1].descr );
}

void interfaceIss_port_description_default( int port )
{
    char pn[64];

    if( port <= (GE_START-1) )
        sprintf(pn, "fastethernet1/0/%d", port);
    else
        sprintf(pn, "gigabitethernet1/0/%d", port);
    strcpy( ifIssEntry[port-1].descr, pn );
}

void init_ifIssEntry(void)
{
    long type = 6L;
    long mtu = 1500L;
    char physAddress[SPRINT_MAX_LEN];
    long adminStatus = 1L;
    long operStatus = 1L;
    lastChanged = time(NULL);
    char specific[] = ".0.0";
    long int port;
    char pn[64];
    
    memset( (char*)physAddress, 0, ETH_MAC_LEN + 1);
    for( port=BASED_PORT ; port<=MAX_ifIssEntry_rowNumber ; port++ )
    {
        if( port <= (GE_START-1) )
        {
            sprintf(pn, "fastethernet1/0/%d", port);
            append_ifIssEntry(port, pn, type, mtu, 
                0L, physAddress, adminStatus, operStatus, lastChanged,
                0L, 0L, 0L, 0L, 0L, 0L,  0L, 0L, 0L, 0L, 0L, 0L, specific);
        }
        else
        {
            sprintf(pn, "gigabitethernet1/0/%d", port);
            append_ifIssEntry(port, pn, type, mtu, 
                0L, physAddress, adminStatus, operStatus, lastChanged,
                0L, 0L, 0L, 0L, 0L, 0L,  0L, 0L, 0L, 0L, 0L, 0L, specific);
        }
    }
}

int append_ifIssEntry(long index,  char *descr,  long type,  long mtu,  u_long speed,
    char *physAddress, long adminStatus, long operStatus, u_long lastChange,
    u_long inOctets, u_long inUCastPkts, u_long inNUCastPkts,
    u_long inDiscards, u_long inErrors, u_long inUnknownProtos,
    u_long outOctets, u_long outUCastPkts, u_long outNUCastPkts,
    u_long outDiscards, u_long outErrors, u_long outQLen, char  *specific)
{
    int row;
    oid outOid[MAX_OID_LEN];
    u_long oidLen;
   
    if (ifIssEntry_rowNumber > MAX_ifIssEntry_rowNumber)
        return -1;

    row = ifIssEntry_rowNumber;
    memset((char*)&ifIssEntry[row], 0, sizeof(struct MIB_ifIssEntry));
   	
    ifIssEntry[row].index = index;

    strcpy (ifIssEntry[row].descr, descr);
    ifIssEntry[row].descr_len = strlen(descr);

    ifIssEntry[row].type = type;
    ifIssEntry[row].mtu = mtu;
    ifIssEntry[row].speed = speed;

    memcpy(ifIssEntry[row].physAddress, physAddress, ETH_MAC_LEN);
    ifIssEntry[row].physAddress_len = ETH_MAC_LEN;

    ifIssEntry[row].adminStatus = adminStatus;
    ifIssEntry[row].operStatus = operStatus;
    ifIssEntry[row].lastChange = lastChange;

    ifIssEntry[row].inOctets = inOctets;
    ifIssEntry[row].inNUCastPkts = inNUCastPkts;
    ifIssEntry[row].inUCastPkts = inUCastPkts;
    ifIssEntry[row].inDiscards = inDiscards;
    ifIssEntry[row].inErrors = inErrors;
    ifIssEntry[row].inUnknownProtos = inUnknownProtos;

    ifIssEntry[row].outOctets = outOctets;
    ifIssEntry[row].outUCastPkts = outUCastPkts;
    ifIssEntry[row].outNUCastPkts = outNUCastPkts;
    ifIssEntry[row].outDiscards = outDiscards;
    ifIssEntry[row].outErrors = outErrors;
    ifIssEntry[row].outQLen = outQLen;

    oidLen = parse_miboid(specific, &outOid[0]);
    memcpy(ifIssEntry[row].specific, outOid, oidLen*sizeof(oid));
    ifIssEntry[row].specific_len = oidLen;

    ifIssEntry_rowNumber++;
    return 0;
}

unsigned char *var_interfacesIss( struct variable *vp, oid *name, 
                                  size_t *length, int exact,
                                  size_t  *var_len, WriteMethod **write_method)
{
    static long ulong_ret;

    if (header_generic(vp, name, length, exact, var_len, write_method) == MATCH_FAILED )
        return NULL;

    switch(vp->magic)
    {
        case IFIssNUMBER:
            ulong_ret = ifIssScalar.ifIssNumber;
            return (u_char*) &ulong_ret;
        default:
            ERROR_MSG("var_interfacesIss() Fail");
    }
    return NULL;
}

struct MIB_ifIssEntry *get_ifIssEntry(u_long index)
{
    long curr_index_value, row;

    curr_index_value = index;
    for( row=0 ; row < ifIssEntry_rowNumber ; row++ )
    {
        if( curr_index_value == ifIssEntry[row].index )
            break;
    }
	
    if( row == ifIssEntry_rowNumber )
        return NULL;

    return (struct MIB_ifIssEntry *)&ifIssEntry[row];
}

struct MIB_ifIssEntry *getnext_ifIssEntry(u_long index)
{
    long curr_index_value, row;

    curr_index_value = index;
    for (row = 0; row < ifIssEntry_rowNumber; row++)
    {
        if (curr_index_value < ifIssEntry[row].index)
            break;
    }

    if( row == ifIssEntry_rowNumber )
        return NULL;

    return (struct MIB_ifIssEntry*)&ifIssEntry[row];
}

unsigned char *var_ifIssTable( struct variable *vp, oid *name,
                               size_t *length, int exact,
                               size_t *var_len, WriteMethod **write_method)
{
    u_long index;
    struct MIB_ifIssEntry *geptr;
    static u_long ulong_ret=0;
    static oid objid[MAX_OID_LEN];
    u_long oidLen;
    char mac[6];
    snmp_eventMsg_t event;

 /* To write value to database. */
    switch( vp->magic )
    {
        case IFIssADMINSTATUS:
            *write_method = write_ifIssAdminStatus;
            break;
        default:
            *write_method = NULL;
            break;
    }

    if( memcmp(name, vp->name, sizeof(oid) * vp->namelen) != 0 )
    {
        memcpy(name, vp->name, sizeof(oid) * vp->namelen);
        *length = vp->namelen;
    }

 /* oid of length is 10, 1,3,6,1,2,1,2,2,1,1,[interface number]. */
 /* the value of exact means it has interface number or not. */
    if( exact )
    {
        if( *length != 11 )
            return NULL;

        index = name[10];
        geptr = get_ifIssEntry(index);
    }
    else
    {
        index = *length > 10 ? name[10] : 0;
        geptr = getnext_ifIssEntry(index);
        if( geptr )
        {
            name[10] = geptr->index;
      	    *length = 11;
        }
        index++;
    }

    if( geptr == NULL )
        return NULL;

    *var_len = sizeof(ulong_ret);
    memset( &event, 0, sizeof(snmp_eventMsg_t) );
    event.port=(unsigned int)index;
    switch( vp->magic )
    {
        case IFIssINDEX:
            ulong_ret = geptr->index;
            return (u_char *) &ulong_ret;

        case IFIssDESCR:
            *var_len = strlen(geptr->descr);
            return (u_char *)geptr->descr;

        case IFIssTYPE:
            ulong_ret = geptr->type;
            return (u_char *) &ulong_ret;

        case IFIssMTU:
            ulong_ret = geptr->mtu;
            return (u_char *) &ulong_ret;

        case IFIssSPEED:
            //ulong_ret = geptr->speed;
            event.type=e_snmp_ifSpeed_get;
            util_snmp_evtMsg_ack_handler( &event );            
            ulong_ret = event.val;
            return (u_char *) &ulong_ret;

        case IFIssPHYSADDRESS:
            event.type=e_snmp_ifPhysAddress_get;
            util_snmp_evtMsg_ack_handler( &event );            
            memcpy(geptr->physAddress, event.string, ETH_MAC_LEN);
            geptr->physAddress_len = ETH_MAC_LEN;
            *var_len = geptr->physAddress_len;
            return (u_char *)&geptr->physAddress;

        case IFIssADMINSTATUS:
            event.type=e_snmp_ifAdminStatus_get;
            util_snmp_evtMsg_ack_handler( &event );
            switch( event.val )
            {
                case CST_ENABLE:
                    geptr->adminStatus = 1;
                    break;
                case CST_DISABLE:
                    geptr->adminStatus = 2;
                    break;
                default:
                    geptr->adminStatus = 3;
                    break;
            }
            ulong_ret = (long)geptr->adminStatus;
            return (u_char *) &ulong_ret;

        case IFIssOPERSTATUS:
            event.type=e_snmp_ifOperStatus_get;
            util_snmp_evtMsg_ack_handler( &event );
            switch( event.val )
            {
                case 0:
                    geptr->operStatus = 2;
                    break;
                default:
                    geptr->operStatus = 1;
                    break;
            }
            ulong_ret = (long)geptr->operStatus;
            return (u_char *) &ulong_ret;

        case IFIssLASTCHANGE:
            event.type=e_snmp_ifLastChange_get;
            util_snmp_evtMsg_ack_handler( &event );
            ulong_ret=event.val;
		    return (u_char *) &ulong_ret;

        case IFIssINOCTETS:
            event.type=e_snmp_ifInOctets_get;
            util_snmp_evtMsg_ack_handler( &event );
            ulong_ret=event.val;
		    return (u_char *) &ulong_ret;

        case IFIssINUCASTPKTS:
            event.type=e_snmp_ifInUcastPkts_get;
            util_snmp_evtMsg_ack_handler( &event );
            ulong_ret=event.val;
		    return (u_char *) &ulong_ret;

        case IFIssINNUCASTPKTS:
            event.type=e_snmp_ifInNUcastPkts_get;
            util_snmp_evtMsg_ack_handler( &event );
            ulong_ret=event.val;
		    return (u_char *) &ulong_ret;

        case IFIssINDISCARDS:
            event.type=e_snmp_ifInDiscards_get;
            util_snmp_evtMsg_ack_handler( &event );
            ulong_ret=event.val;
		    return (u_char *) &ulong_ret;

        case IFIssINERRORS:
            event.type=e_snmp_ifInErrors_get;
            util_snmp_evtMsg_ack_handler( &event );
            ulong_ret=event.val;
		    return (u_char *) &ulong_ret;

        case IFIssINUNKNOWNPROTOS:
            event.type=e_snmp_ifInUnknownProtos_get;
            util_snmp_evtMsg_ack_handler( &event );
            ulong_ret=event.val;
		    return (u_char *) &ulong_ret;

        case IFIssOUTOCTETS:
            event.type=e_snmp_ifOutOctets_get;
            util_snmp_evtMsg_ack_handler( &event );
            ulong_ret=event.val;
		    return (u_char *) &ulong_ret;

        case IFIssOUTUCASTPKTS:
            event.type=e_snmp_ifOutUcastPkts_get;
            util_snmp_evtMsg_ack_handler( &event );
            ulong_ret=event.val;
		    return (u_char *) &ulong_ret;

        case IFIssOUTNUCASTPKTS:
            event.type=e_snmp_ifOutNUcastPkts_get;
            util_snmp_evtMsg_ack_handler( &event );
            ulong_ret=event.val;
		    return (u_char *) &ulong_ret;

        case IFIssOUTDISCARDS:
            event.type=e_snmp_ifOutDiscards_get;
            util_snmp_evtMsg_ack_handler( &event );
            ulong_ret=event.val;
		    return (u_char *) &ulong_ret;

        case IFIssOUTERRORS:
            event.type=e_snmp_ifOutErrors_get;
            util_snmp_evtMsg_ack_handler( &event );
            ulong_ret=event.val;
		    return (u_char *) &ulong_ret;

        case IFIssOUTQLEN:
            event.type=e_snmp_ifOutQLen_get;
            util_snmp_evtMsg_ack_handler( &event );
            ulong_ret=event.val;
		    return (u_char *) &ulong_ret;

        case IFIssSPECIFIC:
            oidLen = geptr->specific_len;
            *var_len = oidLen*sizeof(oid);
            memcpy(objid, geptr->specific, oidLen*sizeof(oid));
            return (u_char *)objid;

        default:
            ERROR_MSG("var_ifIssTable() Type Error!");
            break;
    }
    return NULL;
}

int write_ifIssAdminStatus( int action, u_char *var_val, u_char var_val_type, 
                            size_t var_val_len, u_char *statP, 
                            oid *name, size_t name_len)
{
    long newValue = *((long*) var_val);;
    static long oldValue;
    struct MIB_ifIssEntry *geptr = NULL;
    u_long index;
    int port_num; 
    int bigsize = SNMP_MAX_LEN;
    int val,ret;
    char tmpname[128];
    snmp_eventMsg_t event;
    
    if (action == RESERVE1) {
        if (var_val_type != ASN_INTEGER){
            fprintf(stderr, "write to interfacesIss not ASN_INTEGER\n");
            return SNMP_ERR_WRONGTYPE;
        }
        
        if (var_val_len > sizeof(long)){
            fprintf(stderr,"write to interfacesIss: bad length\n");
            return SNMP_ERR_WRONGLENGTH;
        }
    
        if (newValue < 1L || newValue > 2L) {
            fprintf(stderr,"write to interfacesIss: bad value: %d\n", (int)newValue);
            return SNMP_ERR_WRONGVALUE;
        }
    }

    // 1.3.6.1.2.1.2.2.1.7. [index]
    //                   10   11
    index = name[name_len - 1];
    port_num = (unsigned int)(index);
  
    if( (geptr = get_ifIssEntry(index)) == NULL )
        return SNMP_ERR_INCONSISTENTNAME;
    else
    {
        oldValue = geptr->adminStatus;
 	    geptr->adminStatus = newValue;
        oldValue = geptr->operStatus;
 	    geptr->operStatus = newValue;
        switch (newValue) 
        {
            case 1L:   
                val = CST_ENABLE;
                break;
                
            case 2L:   
                val = CST_DISABLE;
                break;
                
            case 3L:   
            default:
                return SNMP_ERR_WRONGVALUE;
                break;
        }
        event.type = e_snmp_ifAdminStatus_set;
        event.val = val;
        event.port = port_num;
        fprintf(stderr,"%s, %d, %d, %d\n",__func__, event.type, event.port, event.val);
        if( util_snmp_evtMsg_handler(&event) )
            return SNMP_ERR_WRONGVALUE;
    }

    if( action == FREE )
    {
        index = name[10];
        if( (geptr = get_ifIssEntry(index)) != NULL )
            geptr->adminStatus = oldValue;
    }
    return SNMP_ERR_NOERROR;
}
