/******************************************************************************
 *  Copyright 2009 VOLKTEK CORPORATION All rights reserved.
 *
 *  This is unpublished proprietary source code of VOLKTEK CORPORATION
 ******************************************************************************
 */
/*=============================================================================
 * MODULE NAME: The snmp domain module
 * FILE NAME: net_snmp_handle.c
 * PURPOSE: The file provide to parse/handle snmp domain/socket.
 * NOTES: 
 * 
 *
 * AUTHOR: $ 08/09/2009 By Abble Fu $
 *      Create, start coding, and finish first version.
 *
 * MODIFICTION HISTORY:
 * DATE By WHO
 * [INTENTION of MODIFICATION] 
 *   <> - DETAIL DESCRIPTION.
 *       
 *=============================================================================
 */
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/socket.h> 
#include <fcntl.h> 
#include <sys/select.h>  
#include <netinet/in.h> 
#include <pthread.h>
#include <malloc.h> 
#include <linux/netlink.h> 
#include <sys/types.h>
#include <unistd.h>
#include <net-snmp/net-snmp-config.h>
#include <net-snmp/net-snmp-includes.h>
#include <net-snmp/library/snmpUDPDomain.h>
#include <net-snmp/library/vacm.h>
#include <net-snmp/agent/agent_trap.h>
#include <syslog.h>
#include "vtk.h"
#include "net_snmp_handle.h"

#define VTK_ERR_CMD_TYPE                -1      /* Command type error! */
#define VTK_ERR_SNMP_DELETE_TRAP        -180
#define VTK_ERR_SNMP_NO_TRAP            -181
#define VTK_ERR_SNMP_DELETE_COMMUNITY   -182
#define VTK_ERR_SNMP_SET_COMMUNITY      -183
#define VTK_ERR_SNMP_NO_COMMUNITY       -184
#define VTK_ERR_SNMP_SET_TRUST_HOST     -185
#define VTK_ERR_SNMP_NO_VIEW_ENTRY      -186
#define VTK_ERR_SNMP_VIEW_ENTRY_OID     -187
#define VTK_ERR_SNMP_CREATE_VIEW_ENTRY  -188

/* ---------------------------------------------------------------------------
 * Macro Definition
 * ---------------------------------------------------------------------------
 */
static int net_snmp_dbg_flag = 1;
#define _NET_SNMP_DBG(fmt,arg...)   if (net_snmp_dbg_flag ==1)   printf(fmt,##arg)
#define DEFAULT_COMMUNITY   "public"
#define MAX_ARGS 20
#define DBG_COMMENT_CHAR        ';'
#define DIM(x) sizeof(x)/sizeof(x[0])

/* ---------------------------------------------------------------------------
 * Local Function declaration
 * ---------------------------------------------------------------------------
 */
static void *snmp_event_handler(void *raymond);

/* ---------------------------------------------------------------------------
 * Local Variable declaration
 * ---------------------------------------------------------------------------
 */
static int trap_sink_num;
static int vcmsh_serv_config_done = 0;  /* 1: finish load vcmsh server config file */
static int startup_type = 0;            /* 1: warm start */
static int startup_trap_flag = 0;

/* ---------------------------------------------------------------------------
 * External declaration
 * ---------------------------------------------------------------------------
 */
extern char *snmp_trapcommunity;
extern struct trap_sink *sinks;

/* ---------------------------------------------------------------------------
 * Global Variable declaration
 * ---------------------------------------------------------------------------
 */
pthread_t snmp_message_receiver, snmp_event_receiver;
int net_snmp_pid =0;
int snmp_netlink_sock;

/* ---------------------------------------------------------------------------
 * Local Function section
 * ---------------------------------------------------------------------------
 */
int trap_host_delete(char *stream_data,int version)
{
    char *delim=" \t\n";
    char *state_ptr;
    char *ptr ;
    char    *target_sink ;
    char    *s;
    u_short sink_port = 0;
    int ret = -1;
    
    s = calloc(128,1);
    if( s == NULL )
        return -1;

    ptr = strtok_r(stream_data,delim,&state_ptr);
    _NET_SNMP_DBG("%s, %d, %s\n", __func__, __LINE__, ptr);         
    
    if(ptr!=0)
    {
        _NET_SNMP_DBG("%s, %d, \n", __func__, __LINE__); 
        //parse trap sink
        target_sink = ptr;
        ptr = strtok_r(NULL,delim,&state_ptr);
    }
    if(ptr!=0)
    {
        _NET_SNMP_DBG("%s, %d, \n", __func__, __LINE__); 
        //parse trap sink port, if needed
        sink_port = atoi(ptr);
    }
    _NET_SNMP_DBG("%s, %d, \n", __func__, __LINE__); 
    if( sink_port == 0)
    {
        sink_port = 162;
    }
    _NET_SNMP_DBG("%s, %d, \n", __func__, __LINE__); 
    sprintf(s,"udp:%s",target_sink);
    _NET_SNMP_DBG("%s, %d, %s\n", __func__, __LINE__, s); 
    if(version == -1)
    {
        _NET_SNMP_DBG("%s, %d, \n", __func__, __LINE__); 
        if( remove_trap_session2(s,0) == 0 )
        {
            free(s);
            return 0;
        }
        else if(remove_trap_session2(s,1) == 0)
        {
            free(s);
            return 0;
        }
    }
    else
    {
        _NET_SNMP_DBG("%s, %d, %s,ver=%d\n", __func__, __LINE__, s,version); 
        if( remove_trap_session2(s,version) == 0 )
        {
            _NET_SNMP_DBG("%s, %d, \n", __func__, __LINE__); 
            free(s);
            return 0;
        }
    }
    
    free(s);
    return -1;
}

/*-----------------------------------------------------------------------------
 * FUNCTION NAME: snmp_message_handler
 * PURPOSE: This function handle messages from vcmsh'SASI event.
 *
 * PARAMs:
 *      INPUT:  None
 *      OUTPUT: None.
 * RETURN:  None
 *
 * NOTES: 
 *-----------------------------------------------------------------------------
 */
static void *snmp_message_handler(void *raymond) 
{ 
    int  idx; 
    int  comm_fd;
    int  fromlen=0, size, ret,ret1;
    char mode[16], string[NET_SNMP_MAX_BUF_LEN];
    struct sockaddr_in server, client; 
    net_snmp_event_t recvbuf;
    
    comm_fd = socket(AF_INET, SOCK_DGRAM, 0); 
    if(comm_fd < 0)
    { 
        printf("open comm_fd socket fail!\n");
        return ; 
    } 

    memset(&server, 0, sizeof(server));
    server.sin_family     = AF_INET; 
    server.sin_addr.s_addr= inet_addr(LOCAL_ADDRESS);
    server.sin_port       = htons(VTK_SNMP_SERVER_PORT);  
    if( bind(comm_fd, (struct sockaddr *)&server, sizeof(server)) ==-1)
    {
        printf("bind comm_fd socket fail!\n");
        close(comm_fd);
        return ; 
    }

    while(1)
    {   
        memset(&recvbuf, 0x0, sizeof(recvbuf));    
        size = sizeof(client);
        recvfrom( comm_fd,&recvbuf,sizeof(net_snmp_event_t),0,
                  (struct sockaddr *)&client,&size);

        if( recvbuf.type >= e_snmp_cnt_max )
        {
            printf("%s, recvbuf.type(%d) error \n", __func__, recvbuf.type); 
            continue;
        }

        recvbuf.ret=RC_OK;
        switch( recvbuf.type )
        {
            case e_snmp_vcmsh_config_done:
                vcmsh_serv_config_done = 1;
                if(recvbuf.value == 1)
                    startup_type = 1;
                _NET_SNMP_DBG( "%s, %d, startup_type=%d \n", __func__, __LINE__,startup_type);
                break;

            case e_snmp_system_location_get:
                system_mib_get_sysLocation(recvbuf.string);
                break;

            case e_snmp_system_location_set:
                system_mib_set_sysLocation(recvbuf.string);
                break;

            case e_snmp_modle_name_set:
                system_mib_set_version_descr(recvbuf.string);
                break;

            case e_snmp_server_community_get:
                memset(string, 0, sizeof(string));
                ret = show_com2sec_entry_with_net(string);
                if(ret == 0)
                {
                    recvbuf.ret=VTK_ERR_SNMP_NO_COMMUNITY;
                }    
                strcpy( recvbuf.string, string);
                break;

            case e_snmp_server_community_set:
                ret = snmp_community_string_add(recvbuf.access_mode, recvbuf.string, recvbuf.netaddress);
                if( ret == -1)
                {
                    recvbuf.ret=VTK_ERR_SNMP_SET_COMMUNITY;
                }
                else if( ret== -2)
                {
                    recvbuf.ret=VTK_ERR_SNMP_SET_TRUST_HOST;
                }
                break;

            case e_snmp_server_community_delete:
                if(snmp_community_string_delete( recvbuf.access_mode,
                                                 recvbuf.string, recvbuf.netaddress)!= 0)
                {
                    recvbuf.ret=VTK_ERR_SNMP_DELETE_COMMUNITY;
                }
                break;

            case e_snmp_trap_sink_total_num_get:
                recvbuf.value = trap_sink_num;
                break;

            case e_snmp_trap_host_get:
                memset(string, 0, sizeof(string));
                if( show_trap_sinks2(string) == 0 )
                    recvbuf.ret=VTK_ERR_SNMP_NO_TRAP;

                strcpy(recvbuf.string, string);
                break;

            case e_snmp_trap_host_set:
            {
                unsigned short sinkport = 162;
                int version, pdutype;

                switch( recvbuf.value )
                {
                    case 1:
                        version = SNMP_VERSION_1;
                        pdutype = SNMP_MSG_TRAP;
                        break;

                    case 2:
                        version = SNMP_VERSION_2c;
                        pdutype = SNMP_MSG_TRAP2;
                        break;
                }

                /* Check trap entry exist or not, 
                   If trap entry exist, delete this entry and add again.
                   If trap entry not exist, add new one.
                  */
                ret = check_trap_sinks(recvbuf.netaddress,SNMP_VERSION_1);
                ret1 = check_trap_sinks(recvbuf.netaddress,SNMP_VERSION_2c);
                if( (ret == 0) || ( ret == 1) || ( ret1 == 1) )
                {
                    if( ret == 1 ) 
                    {
                        if( trap_host_delete(recvbuf.netaddress,SNMP_VERSION_1) == 0 )
                        {
                            trap_sink_num--;
                        }
                        else
                            printf("replace old SNMP Trap entry FAILED!\n");
                    }

                    if( ret1 == 1 ) 
                    {
                        if( trap_host_delete(recvbuf.netaddress,SNMP_VERSION_2c) == 0 )
                        {
                            trap_sink_num--;
                        }
                        else
                            printf("replace old SNMP Trap entry FAILED!\n");
                   }

                 /* overwrite2: */
                    if( trap_sink_num>MAX_NUM_SNMP_TARP_ENTRY )
                    {
                        printf("Maximum trap entry is %d, Can't add new one.\n", MAX_NUM_SNMP_TARP_ENTRY);  
                        break;
                    }

                    if( create_trap_session(recvbuf.netaddress, sinkport, recvbuf.string,version, pdutype) == 1 )
                    {
                        trap_sink_num++;
                    }
                    else
                        printf("%s, %d, Can't add new trap entry.\n", __func__, __LINE__);  
                }
                else if( sinks == NULL )
                {
                    if( create_trap_session(recvbuf.netaddress, sinkport, recvbuf.string,version, pdutype) == 1 )
                    {
                        trap_sink_num++;
                    }
                    else
                        printf("%s, %d, Can't add new trap entry.\n", __func__, __LINE__);  
                }                
                break;
            }

            case e_snmp_trap_host_delete:
                ret = check_trap_sinks(recvbuf.netaddress,SNMP_VERSION_1);
                if( ret==1 )
                {
                    if( trap_host_delete(recvbuf.netaddress,SNMP_VERSION_1) == 0 )
                        trap_sink_num--;
                }

                ret = check_trap_sinks(recvbuf.netaddress,SNMP_VERSION_2c);
                if( ret==1 )
                {
                    if( trap_host_delete(recvbuf.netaddress,SNMP_VERSION_2c) == 0 )
                        trap_sink_num--;
                }
                break;

            case e_snmp_trap_community_get:
                strcpy( recvbuf.string, snmp_trapcommunity );
                break;

            case e_snmp_trap_community_set:
                strcpy( snmp_trapcommunity, recvbuf.string );
                break;

            case e_snmp_system_contact_get:
                system_mib_get_sysContact(recvbuf.string);
                break;

            case e_snmp_system_contact_set:
                system_mib_set_sysContact(recvbuf.string);
                break;

            case e_snmp_system_name_get:
                system_mib_get_sysName(recvbuf.string);
                break;

            case e_snmp_system_name_set:
                system_mib_set_sysName(recvbuf.string);
                break;

            case e_snmp_pid_get:
                if( net_snmp_pid == 0 )
                    net_snmp_pid = getpid();
                
                fromlen = sizeof (net_snmp_pid); 
                _NET_SNMP_DBG("\n%s, %d: net_snmp_pid=%d \n", __func__, __LINE__, net_snmp_pid); 
                idx = net_snmp_pid;
                if( sendto( comm_fd, &idx , sizeof(idx), 0,
                            (struct sockaddr *)&client, sizeof(client)) < 0)
                {
                    printf("%s, sendto error!!\n", __func__); 
                }
                continue;

            case e_snmp_port_description_set:
                interfaceIss_port_description_set(recvbuf.value,recvbuf.string);
                break;

            case e_snmp_port_description_get:
                interfaceIss_port_description_get(recvbuf.value,recvbuf.string);
                break;

            case e_snmp_port_description_default:
                interfaceIss_port_description_default(recvbuf.value);
                break;

            case e_snmpv3_view_get:
                {
                    struct vacm_viewEntry *gp = NULL;
                    //struct snmpv3_vacm_viewEntry *gp = NULL;
                    int i;

                    _NET_SNMP_DBG("%s, %d:\n", __func__, __LINE__);
                    vacm_scanViewInit();
                    gp = vacm_scanViewNext();
#if 0
                    if(gp == NULL)
                    {
                        recvbuf.ret=VTK_ERR_SNMP_NO_VIEW_ENTRY;
                    }    
                    //memcpy( recvbuf.group_view, gp, sizeof(struct snmpv3_vacm_viewEntry));
                    recvbuf.group_view=gp;
#else
                    if (gp == NULL)
                        _NET_SNMP_DBG("Empty SNMPv3 View entry.\n");
                    else 
                    {
                        _NET_SNMP_DBG("%s, %d:\n", __func__, __LINE__);
                        for (; gp!=NULL; gp=gp->next) 
                        {
                            if (strncmp(gp->viewName+1, "anonymousView", strlen("anonymousView")) == 0)
                                continue;
                            _NET_SNMP_DBG("\n");
                            _NET_SNMP_DBG("View Name: %s\n", gp->viewName+1);
                            _NET_SNMP_DBG("View Subtree: ");
                            for (i=1; i<gp->viewSubtreeLen; i++) 
                            {
                                _NET_SNMP_DBG(".%d", gp->viewSubtree[i]);
                            }
                            _NET_SNMP_DBG("\n");
                            if (gp->viewType == SNMP_VIEW_INCLUDED)
                              _NET_SNMP_DBG("View Type: Included\n");
                            else if (gp->viewType == SNMP_VIEW_EXCLUDED)
                                    _NET_SNMP_DBG("View Name: Excluded\n");
                            _NET_SNMP_DBG("\n");
                        }
                    }
#endif
                }
                break;

            case e_snmpv3_view_set:
                {
                    oid suboid[MAX_OID_LEN];
                    size_t suboid_len = MAX_OID_LEN;
                    char cmdLine[SPRINT_MAX_LEN];
                    _NET_SNMP_DBG("%s, %d:\n", __func__, __LINE__);
                    if (!snmp_parse_oid(recvbuf.string, suboid, &suboid_len)) 
                    {
                        printf("%s, %d: Warning: bad SUBTREE object id(%s)!\n", __func__, __LINE__, recvbuf.string); 
                        recvbuf.ret = VTK_ERR_SNMP_VIEW_ENTRY_OID;
                        break;
                    }
                    _NET_SNMP_DBG("%s, %d:\n", __func__, __LINE__);

                    snprintf(cmdLine, sizeof(cmdLine), "%s %s %s %s", recvbuf.netaddress, recvbuf.access_mode, recvbuf.string, "ff");
                    cmdLine[ sizeof(cmdLine)-1 ] = 0;
                    vacm_parse_view("view", cmdLine);

                    sprintf(cmdLine, "snmp view %s %s %s", recvbuf.netaddress, recvbuf.string, recvbuf.access_mode);
                    _NET_SNMP_DBG("%s, %d:\n", __func__, __LINE__);
                    //if (createWriteList(cmdLine) < 0)
                    //{
                    //    printf("%s, %d: Create SNMPv3 view fail!(%s)\n", __func__, __LINE__, cmdLine); 
                    //  recvbuf.ret = VTK_ERR_SNMP_CREATE_VIEW_ENTRY;
                    //}
                }
                break;

            case e_snmpv3_user_get:
                break;

            case e_snmpv3_user_set:
                break;

            case e_snmpv3_group_get:
                break;

            case e_snmpv3_group_set:
                break;

            case e_snmpv3_access_get:
                break;

            case e_snmpv3_access_set:
                break;

            default:
                recvbuf.ret=VTK_ERR_CMD_TYPE;
                break;
        }

        if( sendto( comm_fd, &recvbuf, sizeof(recvbuf), 0, 
                    (struct sockaddr *)&client, sizeof(client)) < 0)
        {
            printf("%s, sendto error!!\n", __func__); 
        }
    } 
} 

/*-----------------------------------------------------------------------------
 * FUNCTION NAME: snmp_event_handler
 * PURPOSE: This function handle event from kernel specified event.
 *
 * PARAMs:
 *      INPUT:  None.
 *      OUTPUT: None.
 * RETURN:  None.
 * NOTES: 
 *-----------------------------------------------------------------------------
 */
static void *snmp_event_handler(void *raymond)
{
    struct sockaddr_nl src_addr;
    struct link_event{
        int port;
        int event;
    };
    char msg[NET_SNMP_MAX_BUF_LEN];
    struct link_event data;
    int ret;

    net_snmp_pid = getpid();
    while(1)
    {   
        /* wait vcmsh server to finish load config file. */
        if(vcmsh_serv_config_done)
            break;

        sleep(1);
    }
       
    /* netlink event sock */
    snmp_netlink_sock = socket(AF_NETLINK, SOCK_RAW , NETLINK_USERSOCK);
    memset(&src_addr, 0, sizeof(src_addr));
    src_addr.nl_family = AF_NETLINK;
    src_addr.nl_pid = getpid();     /* self pid */
    src_addr.nl_groups = 0;         /* not in mcast groups */
    if(bind(snmp_netlink_sock, (struct sockaddr*)&src_addr, sizeof(src_addr)) < 0 )
    {
        printf("bind netlink socket error.\n");
        return;
    }
    
    /* send snmp trap for warmstart or coldstart */
    if(startup_trap_flag==0)
    {
        _NET_SNMP_DBG("%s, %d: wait for port linkup\n", __func__, __LINE__);
        /* wait for port linkup */
        sleep(13);
        if(startup_type==1)
            send_easy_trap(SNMP_TRAP_WARMSTART, 1);
        else
            send_easy_trap(SNMP_TRAP_COLDSTART, 0);

        startup_trap_flag=1;
    }    
    
    while(1)
    {
        memset(msg, 0, NET_SNMP_MAX_BUF_LEN);
        ret = 0;
        ret = recvfrom(snmp_netlink_sock, msg, NET_SNMP_MAX_BUF_LEN, 0, NULL, 0);
        memcpy(&data, msg+sizeof(struct nlmsghdr), sizeof(data));

        if(data.event == 1)
        {
            _NET_SNMP_DBG("%s, %d: P%d SNMP_TRAP_LINKUP.\n", __func__, __LINE__, data.port);
            send_easy_trap(SNMP_TRAP_LINKUP, data.port);
        }
        else if(data.event == 0)
        {
            _NET_SNMP_DBG("%s, %d: P%d SNMP_TRAP_LINKDOWN.\n", __func__, __LINE__, data.port);
            send_easy_trap(SNMP_TRAP_LINKDOWN, data.port);
        }
        else if(data.event == 4)
        {
            _NET_SNMP_DBG("%s, %d: P%d SNMP_TRAP_LOOPGUARD_SHUTPORT.\n", __func__, __LINE__, data.port);
            send_easy_trap(SNMP_TRAP_ENTERPRISESPECIFIC, data.port);
        }
        else if(data.event == 5)
        {
            _NET_SNMP_DBG("%s, %d: P%d SNMP_TRAP_STPTOPOLOGYCHNAGE.\n", __func__, __LINE__, data.port);
            send_easy_trap(SNMP_TRAP_STPTOPOLOGYCHNAGE, 0);
        }
        else
            continue;
    }
}

/* ---------------------------------------------------------------------------
 * Global Function section
 * ---------------------------------------------------------------------------
 */
/*-----------------------------------------------------------------------------
 * FUNCTION NAME: snmp_evtMsgInit
 * PURPOSE: The initial function for net snmp thread
 *
 * PARAMs:
 *      INPUT:  None.
 *      OUTPUT: None.
 *
 * RETURN:  None.
 *
 * NOTES: 
 *-----------------------------------------------------------------------------
 */
void snmp_evtMsgInit(void)
{
    trap_sink_num = 0;

    if(pthread_create(&snmp_message_receiver, NULL, (void *)snmp_message_handler , NULL) == -1)
    {
        printf("%s create pthread failure!!", __func__);
    }       

    if(pthread_create(&snmp_event_receiver, NULL, (void *)snmp_event_handler , NULL) == -1)
    {
        pthread_cancel(snmp_message_receiver);
        printf("%s create pthread failure!!", __func__);
    }       

    /* Randomize for triggered update random(). */
    srand (time (NULL));

    if( snmp_trapcommunity == NULL )
    {
         snmp_trapcommunity = calloc(128,1);
         memset(snmp_trapcommunity,0,128);
    }

     //specify the default snmp trap community string
     strcpy(snmp_trapcommunity,DEFAULT_COMMUNITY);
}
/*------------------------------End of Program--------------------------------*/
