/** @file lldp_debug.h
 *
 * OpenLLDP Debug Header
 *
 * See LICENSE file for more info.
 * 
 * File: lldp_debug.h
 * 
 * Authors: Terry Simons (terry.simons@gmail.com)
 *
 *******************************************************************/

#ifndef LLDP_DEBUG_H
#define LLDP_DEBUG_H

#include <stdio.h>

/* Borrowed from Open1X */
#define DEBUG_NORMAL         0 // 9

#define NUM_DEBUG_CONFIG     0 // 1
#define DEBUG_CONFIG         0x01

#define NUM_DEBUG_STATE      0 // 2
#define DEBUG_STATE          0x02

#define NUM_DEBUG_TLV        0 // 3
#define DEBUG_TLV            0x04

#define NUM_DEBUG_MSAP       0 // 4
#define DEBUG_MSAP           0x08

#define NUM_DEBUG_INT        0 // 5
#define DEBUG_INT            0x10

#define NUM_DEBUG_SNMP       0 // 6
#define DEBUG_SNMP           0x20

#define NUM_DEBUG_EVERYTHING 0 // 7
#define DEBUG_EVERYTHING     0x40

#define NUM_DEBUG_EXCESSIVE  0 // 8
#define DEBUG_EXCESSIVE      0x80


#define RC_ERROR -1
#define IOCTL_MAGIC 0x5EA0
#define SFP_VENDOR_INFO_LEN 16
#define VTK_LLDP_NEIGHBOR_MAX MAXIMUM_PORT 
#define VTK_LLDP_TLV_TYPE_MAX 9	/* [IEE Std 802.1AB-2009]Table 8-1 */

/* Timing Parameters */
#define VTK_LLDP_DEFAULT_TX_INTERVAL	30
#define VTK_LLDP_DEFAULT_TX_HOLD		4
#define VTK_LLDP_DEFAULT_TX_DELAY		2
#define VTK_LLDP_DEFAULT_TX_REINIT		2
#define VTK_LLDP_DEFAULT_TX_TTL			120


/* Borrowed from Open1X */
int logfile_setup(char *);
void logfile_cleanup();
void lowercase(char *);
void debug_setdaemon(int);
void debug_printf(unsigned char, char *, ...);
void debug_printf_nl(unsigned char, char *, ...);
void debug_hex_printf(uint32_t, uint8_t *, int);
void debug_hex_dump(unsigned char, uint8_t *, int);
void debug_hex_strcat(uint8_t *dest, uint8_t *hextodump, int size);
void debug_set_flags(int);
void debug_alpha_set_flags(char *);
int xsup_assert_long(int, char *, int, char *, int, const char *);
int debug_getlevel();
void ufprintf(FILE *fh, char *instr, int level);

#ifndef LLDP_FUNCTION_DEBUG
#define LLDP_FUNCTION_DEBUG debug_printf(DEBUG_EXCESSIVE, "[DEBUG] %s:%d:%s()\n", __FILE__, __LINE__, __FUNCTION__);
#endif

#endif /* LLDP_DEBUG_H */
