/** @file lldp_linux_framer.c
 * 
 * See LICENSE file for more info.
 *
 * Authors: Terry Simons (terry.simons@gmail.com)
 * 
 **/
#include <netinet/in.h>
#include <sys/ioctl.h>
#include <malloc.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <linux/types.h>
#include <linux/if_packet.h>
#include <linux/if_ether.h>
#include <arpa/inet.h>
#include <net/if.h>
#include <unistd.h>
#include <sys/types.h>
#include <stdio.h>
#include <string.h>
#include <strings.h>
#include <errno.h>
#include <time.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "sasi_cmd.h"
#include "lldp_port.h"
#include "lldp_debug.h"
/*
#include "sasi_cmd.h"
#include "lldp_linux_framer.h"
#include "lldp_port.h"
#include "lldp_debug.h"
#include "framehandler.h"
*/

#define XENOSOCK 0
#define RC_ERROR -1
/* -------------------------------------------------------------------------------
	Tina_test
 --------------------------------------------------------------------------------*/
int tina_debug_flag = 0;
#define TINA_DBG(fmt,args...)  \
    if( tina_debug_flag )\
        printf(fmt, ##args)


/* =====================================
 * Notice:
 *          lldpd     (User Space)
 *              ^
 *              |   I/O Control
 *              v
 *          DRV_SASI    (Kernel Spce)
 * ====================================
 */

#define SASI_NODENAME "/dev/sasi-dev"
static int devfd=-1;

/* -------------------------------------------------------------------------------
 * FUNCTION NAME: lldp_io_control_init
 * PURPOSE: To open sasi-dev io device. 
 *
 * PARAMs:
 *		INPUT:	
 *		OUTPUT:
 *
 * RETURN:	
 * NOTES:
 * -------------------------------------------------------------------------------
 */  
int lldp_io_control_init(void)
{
    devfd = open(SASI_NODENAME,O_RDWR);
    TINA_DBG("%s(%d): devfd =%d\n",__func__,__LINE__,devfd );
    return devfd;
}

/* -------------------------------------------------------------------------------
 * FUNCTION NAME: lldp_io_control_close
 * PURPOSE: To close io device.
 *
 * PARAMs:
 *		INPUT:	
 *		OUTPUT:
 *
 * RETURN:	
 * NOTES:
 * -------------------------------------------------------------------------------
 */  
int lldp_io_control_close(void)
{
    if(devfd > 0)
        close(devfd);
    return 0;
}
 
/* -------------------------------------------------------------------------------
 * FUNCTION NAME: lldp_io_control_access
 * PURPOSE: The sasi-dev io device access function.
 *
 * PARAMs:
 *		INPUT:	cmd - io access type.
 *              *access - the data pointer for access 
 *		OUTPUT:
 *
 * RETURN:	
 * NOTES:
 * -------------------------------------------------------------------------------
 */  
int lldp_io_control_access(int cmd, unsigned long *access)
{
    int ret;
	
    if(access == NULL)
        return RC_ERROR;

    //if( lldp_io_control_init() < 0 )
    ret= lldp_io_control_init();
    if(ret<0)
	{
		TINA_DBG("%s(%d): ret =%d\n",__func__,__LINE__,ret ); 
        return RC_ERROR;
	}

    ret = ioctl(devfd, cmd, (void*)access);//SIOCGIFHWADDR : : struct ifreq ifr
    TINA_DBG("%s(%d): ret =%d\n",__func__,__LINE__,ret );
    lldp_io_control_close();
    return ret;
}

/* -------------------------------------------------------------------------------
 * FUNCTION NAME: lldp_cmd_lldp_remore_port_info_get
 * PURPOSE: Get lldp structure.
 *
 * PARAMs:
 *		INPUT:	port - port ID (neighbor from).
 *              *lldp_port - the data pointer for lldp_port sturcture 
 *		OUTPUT:
 *
 * RETURN:	
 * NOTES:
 * -------------------------------------------------------------------------------
 */
int lldp_cmd_lldp_state_get(int *val)  
{
    sasi_cmd_t cmd;

    cmd.maintype=sasicmd_lldp;
    cmd.subtype=e_lldp_cmd_state_get;
    cmd.portno=0;
    lldp_io_control_access(sasicmd_lldp, (unsigned long *)&cmd); 
    *val = cmd.lldp.state;
    return cmd.ret; 
} 

/* -------------------------------------------------------------------------------
 * FUNCTION NAME: lldp_cmd_lldp_remote_port_info_set
 * PURPOSE: Set lldp structure.
 *
 * PARAMs:
 *		INPUT:	port - port ID (neighbor from).
 *              *lldp_port - the data pointer for lldp_port sturcture 
 *		OUTPUT:
 *
 * RETURN:	
 * NOTES:
 * -------------------------------------------------------------------------------
 */
int lldp_cmd_lldp_remote_port_info_set(int port, sasi_cmd_lldp_t *lldp_cfg)  
{
    sasi_cmd_t cmd;
     
    cmd.maintype=sasicmd_lldp;
    cmd.subtype=e_lldp_cmd_remote_info_set;
    cmd.portno=port;
    memcpy( &cmd.lldp, lldp_cfg, sizeof(sasi_cmd_lldp_t) );     
    lldp_io_control_access(sasicmd_lldp, (unsigned long *)&cmd);
    return cmd.ret;
} 

/* -------------------------------------------------------------------------------
 * FUNCTION NAME: e_lldp_cmd_remote_info_clean
 * PURPOSE: Clean lldp structure.
 *
 * PARAMs:
 *		INPUT:	port - port ID (neighbor from).
 *		OUTPUT:
 *
 * RETURN:	
 * NOTES:
 * -------------------------------------------------------------------------------
 */
int lldp_cmd_lldp_remote_port_info_clean(int port)  
{
    sasi_cmd_t cmd;
     
    cmd.maintype=sasicmd_lldp;
    cmd.subtype=e_lldp_cmd_remote_info_clean; 
    cmd.portno=port;    
    lldp_io_control_access(sasicmd_lldp, (unsigned long *)&cmd);
    return cmd.ret;
} 
    
/* -------------------------------------------------------------------------------
 * FUNCTION NAME: lldp_cmd_lldp_remore_port_info_get
 * PURPOSE: Get lldp structure.
 *
 * PARAMs:
 *		INPUT:	port - port ID (neighbor from).
 *              *lldp_port - the data pointer for lldp_port sturcture 
 *		OUTPUT:
 *
 * RETURN:	
 * NOTES:
 * -------------------------------------------------------------------------------
 */
int lldp_cmd_lldp_remore_port_info_get(int port, sasi_cmd_lldp_t *lldp_cfg)  
{
    sasi_cmd_t cmd;

    cmd.maintype=sasicmd_lldp;
    cmd.subtype=e_lldp_cmd_remote_info_get;
    cmd.portno=port;
    lldp_io_control_access(sasicmd_lldp, (unsigned long *)&cmd); 
    memcpy( lldp_cfg, &cmd.lldp, sizeof(sasi_cmd_lldp_t) );  
    return cmd.ret; 
} 

/* -------------------------------------------------------------------------------
 * FUNCTION NAME: lldp_cmd_lldp_local_port_info_get
 * PURPOSE: Get local port's lldp structure.
 *
 * PARAMs:
 *		INPUT:	port - port ID (neighbor from).
 *              *lldp_port - the data pointer for lldp_port sturcture 
 *		OUTPUT:
 *
 * RETURN:	
 * NOTES:
 * -------------------------------------------------------------------------------
 */
int lldp_cmd_lldp_local_port_info_get(int port, sasi_cmd_lldp_t *lldp_cfg) 
{
    sasi_cmd_t cmd;

    cmd.maintype=sasicmd_lldp;
    cmd.subtype=e_lldp_cmd_local_info_get;    
    cmd.portno=port;
    lldp_io_control_access(sasicmd_lldp, (unsigned long *)&cmd); 
    memcpy( lldp_cfg, &cmd.lldp, sizeof(sasi_cmd_lldp_t) ); 
    return cmd.ret; 
} 

/* -------------------------------------------------------------------------------
 * FUNCTION NAME: lldp_cmd_lldp_port_link_status_get
 * PURPOSE: Get port link status.
 *
 * PARAMs:
 *		INPUT:	port - port ID (neighbor from).
 *              *lldp_port - the data pointer for lldp_port sturcture 
 *		OUTPUT:
 *
 * RETURN:	
 * NOTES:
 * -------------------------------------------------------------------------------
 */
int lldp_cmd_lldp_port_link_status_get(unsigned long *portlist)  
{
    sasi_cmd_t cmd;

    cmd.maintype=sasicmd_lldp;
    cmd.subtype=e_lldp_cmd_local_linkstlist_get;     
    cmd.portno=0;
    lldp_io_control_access(sasicmd_lldp, (unsigned long *)&cmd); 
    *portlist = cmd.lldp.linkstlist;
    return cmd.ret; 
} 

/* -------------------------------------------------------------------------------
 * FUNCTION NAME: lldp_cmd_lldp_port_adminState_get
 * PURPOSE: Get port admin state.
 *
 * PARAMs:
 *		INPUT:	port - port ID (neighbor from).
 *              *lldp_port - the data pointer for lldp_port sturcture 
 *		OUTPUT:
 *
 * RETURN:	
 * NOTES:
 * -------------------------------------------------------------------------------
 */
int lldp_cmd_lldp_port_adminState_get( int port, int *adminState)
{
    sasi_cmd_t cmd;

    cmd.maintype=sasicmd_lldp;
    cmd.subtype=e_lldp_cmd_port_adminState_get;
    cmd.portno=port;
    lldp_io_control_access(sasicmd_lldp, (unsigned long *)&cmd);
    *adminState = cmd.lldp.state;
    return cmd.ret;
}

int lldp_cmd_port_linkStatus_get( int port, int *link)
{
    sasi_cmd_t cmd;

    cmd.maintype=sasicmd_lldp;
    cmd.subtype=e_lldp_cmd_port_linkStatus_get;
    cmd.portno=port;
    lldp_io_control_access(sasicmd_lldp, (unsigned long *)&cmd);
    *link = cmd.lldp.state;
    return cmd.ret;
}

/* -------------------------------------------------------------------------------
 * FUNCTION NAME: lldp_cmd_timing_parameters_get
 * PURPOSE: Get LLDP time parameter.
 *
 * PARAMs:
 *		INPUT:	
 *              
 *		OUTPUT:
 *
 * RETURN:	
 * NOTES:
 * -------------------------------------------------------------------------------
 */
int lldp_cmd_timing_parameters_get( int type, int *value ) 
{
    sasi_cmd_t cmd;

    cmd.maintype=sasicmd_lldp;
    cmd.subtype=e_lldp_cmd_timing_parameters_get;
    cmd.portno=0;
    cmd.lldp.timing_index = type;
    lldp_io_control_access(sasicmd_lldp, (unsigned long *)&cmd);
    *value = cmd.lldp.timing_value;
    return cmd.ret;
}

ssize_t lldp_write(struct lldp_port *lldp_port)
{
	int ret;
	/* Write the frame to the wire */	
	//debug_printf(DEBUG_NORMAL, "[TINA] ldp_port->tx_port= %d\n",lldp_port->tx_port);
	lldp_port->tx.frame[lldp_port->tx.sendsize]=lldp_port->if_index;    //tina  
	lldp_port->tx.sendsize++;  //tina  
	ret = sendto(lldp_port->socket, lldp_port->tx.frame, lldp_port->tx.sendsize, 0, NULL, 0);  
    lldp_port->tx.sendsize--;
    return ret;
}

ssize_t lldp_read(struct lldp_port *lldp_port)
{
    lldp_port->rx.recvsize = recvfrom(lldp_port->socket, lldp_port->rx.frame, lldp_port->mtu, 0, 0, 0); 
	lldp_port->if_index= lldp_port->rx.frame[lldp_port->rx.recvsize-1]; /* Record the Rx port. */
	lldp_port->rx.recvsize--;    
    debug_printf(DEBUG_NORMAL, "[TINA] lldp_port->if_index= %d\n",lldp_port->if_index);
    return(lldp_port->rx.recvsize);
}

void socketCleanupLLDP(struct lldp_port *lldp_port)
{
    if(lldp_port != NULL)
    {
        if(lldp_port->rx.frame != NULL)
        {   
            debug_printf(DEBUG_INT, "Freeing %s rx framebuffer...\n", lldp_port->if_name);
            free(lldp_port->rx.frame);
            lldp_port->rx.frame = NULL;
        }
        else
            debug_printf(DEBUG_NORMAL, "[ERROR] lldp_port->rx.frame was NULL!\n");

        if(lldp_port->tx.frame != NULL)
        {
            debug_printf(DEBUG_INT, "Freeing %s tx framebuffer...\n", lldp_port->if_name);
            free(lldp_port->tx.frame);
            lldp_port->tx.frame = NULL;
        }
        else
            debug_printf(DEBUG_NORMAL, "[ERROR] lldp_port->tx.frame was NULL!\n");

        if(lldp_port->if_name != NULL)
        {
            debug_printf(DEBUG_INT, "Freeing %s...\n", lldp_port->if_name);
            free(lldp_port->if_name);
            lldp_port->if_name = NULL;
        }
        else
            debug_printf(DEBUG_NORMAL, "[ERROR] lldp_port->if_name was NULL!\n");
    }
    else
        debug_printf(DEBUG_NORMAL, "[ERROR] lldp_port was NULL!\n");
}



int socketInitializeLLDP_vtk(int *socket_ret, int *mtu)
{
    struct ifreq *ifr = calloc(1, sizeof(struct ifreq));
    struct sockaddr_ll *sll = calloc(1, sizeof(struct sockaddr_ll));
    int retval             = 0;


    // Set up the raw socket for the LLDP ethertype
    *socket_ret = socket(PF_PACKET, SOCK_RAW, htons(0x88CC)); 
    if(*socket_ret < 0)
    {
        return XENOSOCK;
    }

    // Set up the interface for binding
    sll->sll_family = PF_PACKET;
    sll->sll_ifindex = 1;  
    sll->sll_protocol = htons(0x88CC);

    // Bind the socket and the interface together to form a useable socket:
    retval = bind(*socket_ret, (struct sockaddr *)sll, sizeof(struct sockaddr_ll));
    if(retval < 0)
    {
        debug_printf(DEBUG_NORMAL, "Error binding raw socket to interface %s in %s():%d!\n", *socket_ret, __FUNCTION__, __LINE__);
        return XENOSOCK;
    }

    ifr->ifr_ifindex = 1;  
    strncpy(ifr->ifr_name, "eth0", IF_NAMESIZE);
    if(strlen(ifr->ifr_name) == 0)
    {
        debug_printf(DEBUG_NORMAL, "Invalid interface name in %s():%d\n", __FUNCTION__, __LINE__);
        return XENOSOCK;
    }

    //refreshInterfaceData(lldp_port);
    retval = ioctl(*socket_ret, SIOCGIFFLAGS, ifr);
    if (retval == -1)
    {
		;
    }

    if ((ifr->ifr_flags & IFF_UP) == 0)
    {
        //lldp_port->portEnabled = 0;
    } 

    // set allmulti on interface
    // need to devise a graceful way to turn off allmulti otherwise it is left on for the interface when problem is terminated.
    retval = ioctl(*socket_ret, SIOCGIFFLAGS, ifr);
    if (retval == -1)
    {
		;
    }

    ifr->ifr_flags |=  IFF_ALLMULTI; // allmulti on (verified via ifconfig)
    //  ifr.ifr_flags &= ~IFF_ALLMULTI; // allmulti off (I think)

    retval = ioctl(*socket_ret, SIOCSIFFLAGS, ifr);
    if (retval == -1)
    {
		;
    }

    // Discover MTU of our interface.
    retval = ioctl(*socket_ret, SIOCGIFMTU, ifr);
    if(retval < 0) 
    {
        return retval;
    }   

    *mtu = ifr->ifr_ifru.ifru_mtu;

    free(ifr);
    free(sll);
    return 0;
}

int socketInitializeLLDP(struct lldp_port *lldp_port)
{
    struct ifreq *ifr = calloc(1, sizeof(struct ifreq));
    struct sockaddr_ll *sll = calloc(1, sizeof(struct sockaddr_ll));
    int retval             = 0;

    if(lldp_port->if_name == NULL)
    {
        debug_printf(DEBUG_NORMAL, "Got NULL interface in %s():%d\n", __FUNCTION__, __LINE__);
        return XENOSOCK;
    }

    if(lldp_port->if_index <= 0)
    {
        debug_printf(DEBUG_NORMAL, "'%s' does not appear to be a valid interface name!\n", lldp_port->if_name);
        return XENOSOCK;
    }

    debug_printf(DEBUG_INT, "'%s' is index %d\n", lldp_port->if_name, lldp_port->if_index);

    // Set up the raw socket for the LLDP ethertype
    lldp_port->socket = socket(PF_PACKET, SOCK_RAW, htons(0x88CC));
    if(lldp_port->socket < 0)
    {
        debug_printf(DEBUG_NORMAL, "Couldn't initialize raw socket for interface %s!\n", lldp_port->if_name);
        return XENOSOCK;
    }

    // Set up the interface for binding
    sll->sll_family = PF_PACKET;
    sll->sll_ifindex = lldp_port->if_index;
    sll->sll_protocol = htons(0x88CC);

    // Bind the socket and the interface together to form a useable socket:
    retval = bind(lldp_port->socket, (struct sockaddr *)sll, sizeof(struct sockaddr_ll));
    if(retval < 0)
    {
        debug_printf(DEBUG_NORMAL, "Error binding raw socket to interface %s in %s():%d!\n", lldp_port->if_name, __FUNCTION__, __LINE__);
        return XENOSOCK;
    }

    ifr->ifr_ifindex = lldp_port->if_index;
    strncpy(ifr->ifr_name, &lldp_port->if_name[0], strlen(lldp_port->if_name));
    if(strlen(ifr->ifr_name) == 0)
    {
        debug_printf(DEBUG_NORMAL, "Invalid interface name in %s():%d\n", __FUNCTION__, __LINE__);
        return XENOSOCK;
    }

    refreshInterfaceData(lldp_port);
    retval = ioctl(lldp_port->socket, SIOCGIFFLAGS, ifr);
    if (retval == -1)
    {
        debug_printf(DEBUG_NORMAL, "Can't get flags for interface '%s' in %s():%d!\n", lldp_port->if_name, __FUNCTION__, __LINE__);
    }

    if ((ifr->ifr_flags & IFF_UP) == 0)
    {
        debug_printf(DEBUG_INT, "Interface '%s' is down. portEnabled = 0.\n", lldp_port->if_name);  
        lldp_port->portEnabled = 0;
    } 

    // set allmulti on interface
    // need to devise a graceful way to turn off allmulti otherwise it is left on for the interface when problem is terminated.
    retval = ioctl(lldp_port->socket, SIOCGIFFLAGS, ifr);
    if (retval == -1)
    {
        debug_printf(DEBUG_NORMAL, "Can't get flags for interface '%s' in %s():%d!\n", lldp_port->if_name, __FUNCTION__, __LINE__);
    }

    ifr->ifr_flags |=  IFF_ALLMULTI; // allmulti on (verified via ifconfig)
    //  ifr.ifr_flags &= ~IFF_ALLMULTI; // allmulti off (I think)

    retval = ioctl(lldp_port->socket, SIOCSIFFLAGS, ifr);
    if (retval == -1)
    {
        debug_printf(DEBUG_NORMAL, "Can't set flags for interface '%s' in %s():%d!\n", lldp_port->if_name, __FUNCTION__, __LINE__);
    }

    // Discover MTU of our interface.
    retval = ioctl(lldp_port->socket, SIOCGIFMTU, ifr);
    if(retval < 0) 
    {
        debug_printf(DEBUG_NORMAL, "Can't determine MTU for interface '%s' in %s():%d!\n", lldp_port->if_name, __FUNCTION__, __LINE__);
        return retval;
    }   

    lldp_port->mtu = ifr->ifr_ifru.ifru_mtu;
    debug_printf(DEBUG_INT, "[%s] MTU is %d\n", lldp_port->if_name, lldp_port->mtu);

    lldp_port->rx.frame = calloc(1, lldp_port->mtu - 4);
    lldp_port->tx.frame = calloc(1, (lldp_port->mtu - 2));

    if(!lldp_port->rx.frame)
        debug_printf(DEBUG_NORMAL, "[ERROR] Unable to malloc buffer in %s() at line: %d!\n", __FUNCTION__, __LINE__);
    else
        debug_printf(DEBUG_INT, "Created framebuffer for %s at %x\n", lldp_port->if_name, &lldp_port->rx.frame);

    if(!lldp_port->tx.frame)
        debug_printf(DEBUG_NORMAL, "[ERROR] Unable to malloc buffer in %s() at line: %d!\n", __FUNCTION__, __LINE__);
    else
        debug_printf(DEBUG_INT, "Created framebuffer for %s at %x\n", lldp_port->if_name, &lldp_port->tx.frame);

    debug_printf(DEBUG_INT, "Interface (%s) MTU is %d.\n", lldp_port->if_name, lldp_port->mtu);

    free(ifr);
    free(sll);
    return 0;
}

/***********************************************
 * Get the MAC address of an interface
 ***********************************************/
static int _getmac(struct lldp_port *lldp_port)
{
    struct ifreq *ifr = calloc(1, sizeof(struct ifreq));
    int retval = 0;

    ifr->ifr_ifindex = lldp_port->if_index;
    strncpy(ifr->ifr_name, &lldp_port->if_name[0], strlen(lldp_port->if_name));
    retval = ioctl(lldp_port->socket, SIOCGIFHWADDR, ifr);
    if(retval < 0)
    {
        free(ifr);
        return retval;
    }

    memcpy(&lldp_port->source_mac[0], &ifr->ifr_hwaddr.sa_data[0], 6);
    free(ifr);
    return 0;
}

/***********************************************
 * Get the IP address of an interface
 ***********************************************/
static int _getip(struct lldp_port *lldp_port)
{
    struct ifreq *ifr = calloc(1, sizeof(struct ifreq));
    int retval = 0;
  
    strncpy(ifr->ifr_name, &lldp_port->if_name[0], strlen(lldp_port->if_name));

    // get interface ip address
    retval = ioctl(lldp_port->socket, SIOCGIFADDR, ifr);
    if (retval < 0)
    {
        debug_printf(DEBUG_NORMAL, "SIOCGIFADDR returned an error: %d\n", retval);
    
        // set source ip address to 0.0.0.0 on error. 
        memset(&lldp_port->source_ipaddr[0], 0x00, 4);
    
        // we should consider reverting to the MAC address on error instead.
    
        free(ifr);
        return retval;
    }
    else
        memcpy(&lldp_port->source_ipaddr[0], &ifr->ifr_addr.sa_data[2], 4);
    free(ifr);
    return retval;
}

void refreshInterfaceData(struct lldp_port *lldp_port)
{
    _getmac(lldp_port);
    _getip(lldp_port);
}
