/*******************************************************************
 *
 * OpenLLDP TX Statemachine
 *
 * See LICENSE file for more info.
 * 
 * File: tx_sm.c
 * 
 * Authors: Terry Simons (terry.simons@gmail.com)
 *
 *******************************************************************/

#include <stdint.h>
#include <strings.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <string.h>
#include <stdio.h>
#include <sys/types.h>
#include <malloc.h>
#include <stdlib.h>

#include "lldp_port.h"
#include "tlv.h"
#include "tlv_common.h"
#include "lldp_debug.h"
#include "tx_sm.h"
/*
#include "tx_sm.h"
#include "lldp_port.h"
#include "lldp_debug.h"
#include "tlv.h"
#include "tlv_common.h"
#include "framehandler.h"
*/
#define TRUE 1
#define FALSE 0

extern struct lci_s lci;

void mibConstrInfoLLDPDU(struct lldp_port *lldp_port)
{
    struct eth_hdr tx_hdr;
    struct lldp_tlv_list *tlv_list = NULL;
	struct lldp_flat_tlv *tlv      = NULL;
	struct lldp_tlv_list *tmp      = 0;
	uint32_t frame_offset          = 0;

    // refreshInterfaceData(lldp_port);    /* Get source_mac and source_ipaddr. */

    /* As per section 10.3.1, verify the destination and ethertype */
    tx_hdr.dst[0] = 0x01;
    tx_hdr.dst[1] = 0x80;
    tx_hdr.dst[2] = 0xc2;
    tx_hdr.dst[3] = 0x00;
    tx_hdr.dst[4] = 0x00;
    tx_hdr.dst[5] = 0x0e;

    tx_hdr.src[0] = lldp_port->source_mac[0];
    tx_hdr.src[1] = lldp_port->source_mac[1];
    tx_hdr.src[2] = lldp_port->source_mac[2];
    tx_hdr.src[3] = lldp_port->source_mac[3];
    tx_hdr.src[4] = lldp_port->source_mac[4];
    tx_hdr.src[5] = lldp_port->source_mac[5];

    tx_hdr.ethertype = htons(0x88cc);

    // How far into our frame should we offset our copies? */
    frame_offset = 0;

	if(lldp_port->tx.frame != NULL)
	{
		memcpy(&lldp_port->tx.frame[frame_offset], &tx_hdr, sizeof(tx_hdr));
	}

    frame_offset += sizeof(struct eth_hdr);

    /* This TLV *MUST* be first. */
    add_tlv(create_chassis_id_tlv(lldp_port), &tlv_list);

    /* This TLV *MUST* be second. */
    add_tlv(create_port_id_tlv(lldp_port), &tlv_list);

    /* This TLV *MUST* be third. */
    add_tlv(create_ttl_tlv(lldp_port), &tlv_list);

    add_tlv(create_port_description_tlv(lldp_port), &tlv_list);

    add_tlv(create_system_name_tlv(lldp_port), &tlv_list);

    add_tlv(create_system_description_tlv(lldp_port), &tlv_list);

    add_tlv(create_system_capabilities_tlv(lldp_port), &tlv_list);

    add_tlv(create_management_address_tlv(lldp_port), &tlv_list);

    /* add LLDP-MED location identification. */
    if (lci.location_data_format != -1)
        add_tlv(create_lldpmed_location_identification_tlv(lldp_port),&tlv_list);

    /* This TLV *MUST* be last. */
    add_tlv(create_end_of_lldpdu_tlv(lldp_port), &tlv_list);

    tmp = tlv_list;
    while(tmp != NULL)
    {
        tlv = flatten_tlv(tmp->tlv);
		if(lldp_port->tx.frame != NULL)
			memcpy(&lldp_port->tx.frame[frame_offset], tlv->tlv, tlv->size);

        frame_offset += tlv->size;
        destroy_flattened_tlv(&tlv);
        tmp = tmp->next;
    }
    destroy_tlv_list(&tlv_list);

    /* Pad to 64 bytes. */
    if(frame_offset < 64)
        lldp_port->tx.sendsize = 64;
    else
        lldp_port->tx.sendsize = frame_offset;
}

static void mibConstrShutdownLLDPDU(struct lldp_port *lldp_port)
{
    struct lldp_tlv *end_of_lldpdu_tlv = create_end_of_lldpdu_tlv(lldp_port);

	debug_printf(DEBUG_NORMAL, "Would send shutdown!");

    if(validate_end_of_lldpdu_tlv(end_of_lldpdu_tlv))
    {
        // We only want the first 3 bytes... 
        // The type/length pair are bytes 1 & 2 (7 and 9 bits respectively)
        // The 3rd octect should contain a 0.
        memcpy(&lldp_port->tx.frame[0], end_of_lldpdu_tlv, 3);
        debug_printf(DEBUG_NORMAL, "%s(%d): \n", __FUNCTION__, __LINE__);
    }
    else
        debug_printf(DEBUG_NORMAL, "%s(%d): [ERROR] TLV End of LLDPDU validation faliure!\n", __FUNCTION__, __LINE__);//???
    free(end_of_lldpdu_tlv);
}

uint8_t txFrame(struct lldp_port *lldp_port)
{
	int val;
    debug_printf(DEBUG_INT, "Sending LLDPDU on [%s]!\n", lldp_port->if_name);

	lldp_cmd_lldp_state_get(&val); 
    if(val==1)//LLDP enable
    	lldp_write(lldp_port); 

    lldp_port->tx.statistics.statsFramesOutTotal++;
    return 0;
}

uint8_t txInitializeLLDP(struct lldp_port *lldp_port)
{
	int index, ret, time[2];     
    
    /* As per IEEE 802.1AB section 10.1.1 */
    lldp_port->tx.somethingChangedLocal = 0;

    /* Defined in 10.5.2.1 */
    lldp_port->tx.statistics.statsFramesOutTotal = 0;

    lldp_port->tx.timers.msgTxHold     = VTK_LLDP_DEFAULT_TX_HOLD;
    lldp_port->tx.timers.reinitDelay   = VTK_LLDP_DEFAULT_TX_DELAY;
    lldp_port->tx.timers.txDelay       = VTK_LLDP_DEFAULT_TX_REINIT;	
	for(index=0; index<2; index++)
	    ret=lldp_cmd_timing_parameters_get( index, &time[index]); 
    lldp_port->tx.timers.msgTxInterval = time[0];
    lldp_port->tx.txTTL = time[1];
    
    // Unsure what to set these to...
    lldp_port->tx.timers.txShutdownWhile = 0;

    /* Collect all of the system specific information here */
    return 0;
}

static char *txStateFromID(uint8_t state)
{
    switch(state)
    {
        case TX_LLDP_INITIALIZE:
            return "TX_LLDP_INITIALIZE";
        case TX_IDLE:
            return "TX_IDLE";
        case TX_SHUTDOWN_FRAME:
            return "TX_SHUTDOWN_FRAME";
        case TX_INFO_FRAME:
            return "TX_INFO_FRAME";
    }
    debug_printf(DEBUG_NORMAL, "[ERROR] Unknown TX State: '%d'\n", state);
    return "Unknown";
}

static uint16_t min(uint16_t value1, uint16_t value2)
{
    if(value1 < value2)
        return value1;
    return value2;
}

void txChangeToState(struct lldp_port *lldp_port, uint8_t state)
{
    debug_printf(DEBUG_STATE, "%s -> %s\n", txStateFromID(lldp_port->tx.state), txStateFromID(state));
    switch(state)
    {
        case TX_LLDP_INITIALIZE:
            if((lldp_port->tx.state != TX_SHUTDOWN_FRAME) && lldp_port->portEnabled)
                debug_printf(DEBUG_STATE, "%s(%d): Illegal Transition: [%s] %s -> %s\n",__func__,__LINE__, lldp_port->if_name, txStateFromID(lldp_port->tx.state), txStateFromID(state));  
            break;
            
        case TX_IDLE:
            if(!(lldp_port->tx.state == TX_LLDP_INITIALIZE ||
               lldp_port->tx.state == TX_INFO_FRAME))
            {
               debug_printf(DEBUG_STATE, "%s(%d): Illegal Transition: [%s] %s -> %s\n",__func__,__LINE__, lldp_port->if_name, txStateFromID(lldp_port->tx.state), txStateFromID(state));
            }
            
            lldp_port->tx.txTTL = min(65535, (lldp_port->tx.timers.msgTxInterval * lldp_port->tx.timers.msgTxHold));
            lldp_port->tx.timers.txTTR = lldp_port->tx.timers.msgTxInterval;
            lldp_port->tx.somethingChangedLocal = FALSE;
            lldp_port->tx.timers.txDelayWhile = lldp_port->tx.timers.txDelay;
            break;
            
        case TX_SHUTDOWN_FRAME:
        case TX_INFO_FRAME:
            if(lldp_port->tx.state != TX_IDLE)
                debug_printf(DEBUG_STATE, "%s(%d): Illegal Transition: [%s] %s -> %s\n",__func__,__LINE__, lldp_port->if_name, txStateFromID(lldp_port->tx.state), txStateFromID(state));
            break;
            
        default:
            debug_printf(DEBUG_STATE, "%s(%d): Illegal Transition: [%s] %s -> %s\n",__func__,__LINE__, lldp_port->if_name, txStateFromID(lldp_port->tx.state), txStateFromID(state));
    }
    lldp_port->tx.state = state;
}

static void txGlobalStatemachineRun(struct lldp_port *lldp_port)
{
    /* Sit in TX_LLDP_INITIALIZE until the next initialization */    
	debug_printf(DEBUG_NORMAL, "%s(%d): portEnabled=%d\n", __func__, __LINE__,lldp_port->portEnabled);
    if(lldp_port->portEnabled == FALSE)
    {
        lldp_port->portEnabled = TRUE;
        txChangeToState(lldp_port, TX_LLDP_INITIALIZE);
    }

    switch(lldp_port->tx.state)
    {
        case TX_LLDP_INITIALIZE:
            debug_printf(DEBUG_NORMAL, "%s(%d): adminState=%d\n", __func__, __LINE__, lldp_port->adminState);
            if((lldp_port->adminState == lldp_enabledRxTx) || (lldp_port->adminState == lldp_enabledTxOnly))
                txChangeToState(lldp_port, TX_IDLE);            
            break;
            
        case TX_IDLE:  // It's time to send a shutdown frame...
        	debug_printf(DEBUG_NORMAL, "%s(%d): adminState=%d\n", __func__, __LINE__, lldp_port->adminState);
            if( (lldp_port->adminState == lldp_disabled) || (lldp_port->adminState == lldp_enabledRxOnly))
            {
                txChangeToState(lldp_port, TX_SHUTDOWN_FRAME);
                break;
            }
            // It's time to send a frame...            
			debug_printf(DEBUG_NORMAL, "%s(%d): %d, %d, %d\n", __func__, __LINE__,
			lldp_port->tx.timers.txDelayWhile,
			lldp_port->tx.timers.txTTR,
			lldp_port->tx.somethingChangedLocal);
            if((lldp_port->tx.timers.txDelayWhile == 0) && ((lldp_port->tx.timers.txTTR == 0) || (lldp_port->tx.somethingChangedLocal)))
            {
                txChangeToState(lldp_port, TX_INFO_FRAME);
            }
            break;
            
        case TX_SHUTDOWN_FRAME:
            debug_printf(DEBUG_NORMAL, "%s(%d): txShutdownWhile=%d\n", __func__, __LINE__,lldp_port->tx.timers.txShutdownWhile);
            if(lldp_port->tx.timers.txShutdownWhile == 0)
                txChangeToState(lldp_port, TX_LLDP_INITIALIZE);
            break;
            
        case TX_INFO_FRAME:
            debug_printf(DEBUG_NORMAL, "%s(%d): \n", __func__, __LINE__);
            txChangeToState(lldp_port, TX_IDLE);
            break;
        default:
            debug_printf(DEBUG_NORMAL, "[ERROR] The TX State Machine is broken!\n");
    }
}

static void tx_display_timers(struct lldp_port *lldp_port)
{
    debug_printf(DEBUG_STATE, "[IP] (%s) IP: %d.%d.%d.%d\n", lldp_port->if_name, lldp_port->source_ipaddr[0], lldp_port->source_ipaddr[1], lldp_port->source_ipaddr[2], lldp_port->source_ipaddr[3]);
    debug_printf(DEBUG_STATE, "[TIMER] (%s) txTTL: %d\n", lldp_port->if_name, lldp_port->tx.txTTL);
    debug_printf(DEBUG_STATE, "[TIMER] (%s) txTTR: %d\n", lldp_port->if_name, lldp_port->tx.timers.txTTR);
    debug_printf(DEBUG_STATE, "[TIMER] (%s) txDelayWhile: %d\n", lldp_port->if_name, lldp_port->tx.timers.txDelayWhile);
    debug_printf(DEBUG_STATE, "[TIMER] (%s) txShutdownWhile: %d\n", lldp_port->if_name, lldp_port->tx.timers.txShutdownWhile);
}

static void tx_decrement_timer(uint16_t *timer)
{
    if((*timer) > 0)
	{        
    	(*timer)--;
    }
}

static void tx_do_update_timers(struct lldp_port *lldp_port)
{
    tx_decrement_timer(&lldp_port->tx.timers.txShutdownWhile);
    tx_decrement_timer(&lldp_port->tx.timers.txDelayWhile);
    tx_decrement_timer(&lldp_port->tx.timers.txTTR);
    debug_printf(DEBUG_NORMAL, "%s(%d): \n", __FUNCTION__, __LINE__);
    tx_display_timers(lldp_port);
}

static void tx_do_tx_lldp_initialize(struct lldp_port *lldp_port)
{
    /* As per 802.1AB 10.5.4.3 */
    txInitializeLLDP(lldp_port);
}

static void tx_do_tx_idle(struct lldp_port *lldp_port)
{
    /* As per 802.1AB 10.5.4.3 */
    /* I think these belong in the change to state block...
       lldp_port->tx.txTTL = min(65535, (lldp_port->tx.timers.msgTxInterval * lldp_port->tx.timers.msgTxHold));
       lldp_port->tx.timers.txTTR = lldp_port->tx.timers.msgTxInterval;
       lldp_port->tx.somethingChangedLocal = FALSE;
       lldp_port->tx.timers.txDelayWhile = lldp_port->tx.timers.txDelay;
       */
}

static void tx_do_tx_shutdown_frame(struct lldp_port *lldp_port)
{
    /* As per 802.1AB 10.5.4.3 */
    mibConstrShutdownLLDPDU(lldp_port);
    txFrame(lldp_port);
    debug_printf(DEBUG_NORMAL, "%s(%d): reinitDelay = %d\n", __FUNCTION__, __LINE__,lldp_port->tx.timers.reinitDelay);
    lldp_port->tx.timers.txShutdownWhile = 1;//lldp_port->tx.timers.reinitDelay;
    debug_printf(DEBUG_NORMAL, "%s(%d): txShutdownWhile = %d\n", __FUNCTION__, __LINE__,lldp_port->tx.timers.txShutdownWhile);
}

static void tx_do_tx_info_frame(struct lldp_port *lldp_port)
{
    /* As per 802.1AB 10.5.4.3 */
    mibConstrInfoLLDPDU(lldp_port);
    txFrame(lldp_port);
}

void txStatemachineRun(struct lldp_port *lldp_port)
{
    debug_printf(DEBUG_STATE, "%s -> %s\n", lldp_port->if_name, txStateFromID(lldp_port->tx.state));
	//debug_printf(DEBUG_STATE, "%s(%d): %d\n", __func__,__LINE__,lldp_port->portEnabled);
    
    txGlobalStatemachineRun(lldp_port);
    
    switch(lldp_port->tx.state)
    {
        case TX_LLDP_INITIALIZE:
            debug_printf(DEBUG_NORMAL, "%s(%d): \n", __FUNCTION__, __LINE__);
            tx_do_tx_lldp_initialize(lldp_port);
            break;
        case TX_IDLE:
            debug_printf(DEBUG_NORMAL, "%s(%d): \n", __FUNCTION__, __LINE__);
            tx_do_tx_idle(lldp_port);
            break;
        case TX_SHUTDOWN_FRAME:
            debug_printf(DEBUG_NORMAL, "%s(%d): \n", __FUNCTION__, __LINE__);
            tx_do_tx_shutdown_frame(lldp_port);
            break;
        case TX_INFO_FRAME:
            debug_printf(DEBUG_NORMAL, "%s(%d): \n", __FUNCTION__, __LINE__);
            tx_do_tx_info_frame(lldp_port);
            break;
        default:
            debug_printf(DEBUG_NORMAL, "[ERROR] The TX State Machine is broken!\n");
    }
    tx_do_update_timers(lldp_port);
}
