/** @file tx_sm.h
 *
 * \brief OpenLLDP TX Statemachine Header
 *
 * This header file contains declarations, variables, and defines 
 * specific to the OpenLLDP transmit state machine.
 *
 * See LICENSE file for more info.
 * 
 * Authors: Terry Simons (terry.simons@gmail.com)
 *
 *******************************************************************/

#ifndef TX_SM_H
#define TX_SM_H

#include <stdint.h>
#include "lldp_port.h"

// States from the transmit state machine diagram (IEEE 802.1AB 10.5.4.3)
#define TX_LLDP_INITIALIZE 0 /**< Initialize state from IEEE 802.1AB 10.5.4.3 */
#define TX_IDLE            1 /**< Idle state from IEEE 802.1AB 10.5.4.3 */
#define TX_SHUTDOWN_FRAME  2 /**< Shutdown state from IEEE 802.1AB 10.5.4.3 */
#define TX_INFO_FRAME      3 /**< Transmit state from IEEE 802.1AB 10.5.4.3 */
// End states from the trnsmit state machine diagram

/* Defined by the 802.1AB specification */
uint8_t txInitializeLLDP(struct lldp_port *lldp_port);
void mibConstrInfoLLDPDU(struct lldp_port *lldp_port);
uint8_t txFrame();
/* End Defined by the 802.1AB specification */

/* Utility functions not defined in the specification */
void txChangeToState(struct lldp_port *lldp_port, uint8_t state);
void txStatemachineRun(struct lldp_port *lldp_port);
/* End Utility functions not defined in the specification */

#endif /* TX_SM_H */
